# The MIT License (MIT)
# Copyright (c) 2014 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Class for retry options in the Azure Cosmos database service.
"""


from typing import Optional


class RetryOptions:
    """The retry options to be applied to all requests when retrying.

    :ivar int MaxRetryAttemptCount:
        Max number of retries to be performed for a request. Default value 9.
    :ivar int FixedRetryIntervalInMilliseconds:
        Fixed retry interval in milliseconds to wait between each retry ignoring
        the retryAfter returned as part of the response.
    :ivar int MaxWaitTimeInSeconds:
        Max wait time in seconds to wait for a request while the retries are happening.
        Default value 30 seconds.
    """

    def __init__(
        self,
        max_retry_attempt_count: int = 9,
        fixed_retry_interval_in_milliseconds: Optional[int] = None,
        max_wait_time_in_seconds: int = 30
    ):
        self._max_retry_attempt_count = max_retry_attempt_count
        self._fixed_retry_interval_in_milliseconds = fixed_retry_interval_in_milliseconds
        self._max_wait_time_in_seconds = max_wait_time_in_seconds

    @property
    def MaxRetryAttemptCount(self) -> int:
        return self._max_retry_attempt_count

    @property
    def FixedRetryIntervalInMilliseconds(self) -> Optional[int]:
        return self._fixed_retry_interval_in_milliseconds

    @property
    def MaxWaitTimeInSeconds(self) -> int:
        return self._max_wait_time_in_seconds



# The MIT License (MIT)
# Copyright (c) 2014 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Runtime Constants in the Azure Cosmos database service.
"""


class MediaTypes(object):
    """Constants of media types.

    See http://www.iana.org/assignments/media-types/media-types.xhtml for
    more information.
    """

    Any = "*/*"
    ImageJpeg = "image/jpeg"
    ImagePng = "image/png"
    JavaScript = "application/x-javascript"
    Json = "application/json"
    OctetStream = "application/octet-stream"
    QueryJson = "application/query+json"
    SQL = "application/sql"
    TextHtml = "text/html"
    TextPlain = "text/plain"
    Xml = "application/xml"



# The MIT License (MIT)
# Copyright (c) 2014 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Internal class for resource throttle retry policy implementation in the Azure
Cosmos database service.
"""

from . import http_constants


class ResourceThrottleRetryPolicy(object):
    def __init__(self, max_retry_attempt_count, fixed_retry_interval_in_milliseconds, max_wait_time_in_seconds):
        self._max_retry_attempt_count = max_retry_attempt_count
        self._fixed_retry_interval_in_milliseconds = fixed_retry_interval_in_milliseconds
        self._max_wait_time_in_milliseconds = max_wait_time_in_seconds * 1000
        self.current_retry_attempt_count = 0
        self.cumulative_wait_time_in_milliseconds = 0

    def ShouldRetry(self, exception):
        """Returns true if the request should retry based on the passed-in exception.

        :param exceptions.CosmosHttpResponseError exception:
        :returns: a boolean stating whether the request should be retried
        :rtype: bool
        """
        if self.current_retry_attempt_count < self._max_retry_attempt_count:
            self.current_retry_attempt_count += 1
            self.retry_after_in_milliseconds = 0  # pylint: disable=attribute-defined-outside-init

            if self._fixed_retry_interval_in_milliseconds:
                self.retry_after_in_milliseconds = (  # pylint: disable=attribute-defined-outside-init
                    self._fixed_retry_interval_in_milliseconds
                )
            elif http_constants.HttpHeaders.RetryAfterInMilliseconds in exception.headers:
                self.retry_after_in_milliseconds = int(  # pylint: disable = attribute-defined-outside-init
                    exception.headers[http_constants.HttpHeaders.RetryAfterInMilliseconds]
                )

            if self.cumulative_wait_time_in_milliseconds < self._max_wait_time_in_milliseconds:
                self.cumulative_wait_time_in_milliseconds += self.retry_after_in_milliseconds
                return True

        return False



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.
import time

from workload_utils import *
from workload_configs import *

from azure.cosmos import CosmosClient, documents

def run_workload(client_id, client_logger):
    connectionPolicy = documents.ConnectionPolicy()
    connectionPolicy.UseMultipleWriteLocations = USE_MULTIPLE_WRITABLE_LOCATIONS
    with CosmosClient(COSMOS_URI, COSMOS_CREDENTIAL, connection_policy=connectionPolicy,
                      preferred_locations=PREFERRED_LOCATIONS, excluded_locations=CLIENT_EXCLUDED_LOCATIONS,
                      enable_diagnostics_logging=True, logger=client_logger,
                      user_agent=get_user_agent(client_id)) as client:
        db = client.get_database_client(COSMOS_DATABASE)
        cont = db.get_container_client(COSMOS_CONTAINER)
        time.sleep(1)

        while True:
            try:
                upsert_item(cont, REQUEST_EXCLUDED_LOCATIONS, CONCURRENT_REQUESTS)
                read_item(cont, REQUEST_EXCLUDED_LOCATIONS, CONCURRENT_REQUESTS)
                query_items(cont, REQUEST_EXCLUDED_LOCATIONS, CONCURRENT_QUERIES)
            except Exception as e:
                client_logger.info("Exception in application layer")
                client_logger.error(e)


if __name__ == "__main__":
    file_name = os.path.basename(__file__)
    prefix, logger = create_logger(file_name)
    run_workload(prefix, logger)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
# These examples are ingested by the documentation system, and are
# displayed in the SDK reference documentation. When editing these
# example snippets, take into consideration how this might affect
# the readability and usability of the reference documentation.

# All interaction with Cosmos DB starts with an instance of the CosmosClient
# [START create_client]
from azure.cosmos import exceptions, CosmosClient, PartitionKey
from typing import Dict, Any

import os

url = os.environ["ACCOUNT_URI"]
key = os.environ["ACCOUNT_KEY"]
client = CosmosClient(url, key)
# [END create_client]

# Create a database in the account using the CosmosClient,
# specifying that the operation shouldn't throw an exception
# if a database with the given ID already exists.
# [START create_database]
database_name = "testDatabase"
try:
    database = client.create_database(id=database_name)
except exceptions.CosmosResourceExistsError:
    database = client.get_database_client(database=database_name)
# [END create_database]

# Create a container, handling the exception if a container with the
# same ID (name) already exists in the database.
# [START create_container]
container_name = "products"
try:
    container = database.create_container(
        id=container_name, partition_key=PartitionKey(path="/productName")
    )
except exceptions.CosmosResourceExistsError:
    container = database.get_container_client(container_name)
# [END create_container]

# Create a container with custom settings. This example
# creates a container with a custom partition key.
# [START create_container_with_settings]
customer_container_name = "customers"
try:
    customer_container = database.create_container(
        id=customer_container_name,
        partition_key=PartitionKey(path="/city"),
        default_ttl=200,
    )
except exceptions.CosmosResourceExistsError:
    customer_container = database.get_container_client(customer_container_name)
# [END create_container_with_settings]

# Retrieve a container by walking down the resource hierarchy
# (client->database->container), handling the exception generated
# if no container with the specified ID was found in the database.
# [START get_container]
database = client.get_database_client(database_name)
container = database.get_container_client(container_name)
# [END get_container]

# [START list_containers]
database = client.get_database_client(database_name)
for container_dict in database.list_containers():
    print("Container ID: {}".format(container_dict['id']))
# [END list_containers]

# Insert new items by defining a dict and calling Container.upsert_item
# [START upsert_items]
container = database.get_container_client(container_name)
for i in range(1, 10):
    container.upsert_item(
        dict(id="item{}".format(i), productName="Widget", productModel="Model {}".format(i))
    )
# [END upsert_items]

# Modify an existing item in the container
# [START update_item]
item = container.read_item("item2", partition_key="Widget")
item["productModel"] = "DISCONTINUED"
updated_item = container.upsert_item(item)
# [END update_item]

# Query the items in a container using SQL-like syntax. This example
# gets all items whose product model hasn't been discontinued.
# [START query_items]
import json

for queried_item in container.query_items(
    query='SELECT * FROM products p WHERE p.productModel <> "DISCONTINUED"',
    enable_cross_partition_query=True,
):
    print(json.dumps(queried_item, indent=True))
# [END query_items]

# Parameterized queries are also supported. This example
# gets all items whose product model has been discontinued.
# [START query_items_param]
discontinued_items = container.query_items(
    query='SELECT * FROM products p WHERE p.productModel = @model AND p.productName="Widget"',
    parameters=[dict(name="@model", value="DISCONTINUED")],
)
for discontinued_item in discontinued_items:
    print(json.dumps(discontinued_item, indent=True))
# [END query_items_param]

# [START priority_level option]
# Priority-based execution is a capability which allows users to specify priority
# for the request sent to Azure Cosmos DB. Based on the priority specified by the user,
# if there are more requests than the configured RU/s in a second,
# then Azure Cosmos DB will throttle low priority requests to allow high priority requests to execute.
# Can be used for Read, Write, and Query operations. This is specified with the `priority` keyword.
# the value can either be low or high.
for queried_item in container.query_items(
    query='SELECT * FROM products p WHERE p.productModel <> "DISCONTINUED"',
    enable_cross_partition_query=True,
    priority="High"
):
    print(json.dumps(queried_item, indent=True))

# [END priority_level option]

# [START client_level_priority]
# Priority can also be set at the client level, which will apply to all requests made by that client.
# This is useful when you want all operations from a particular client to have the same priority.
# The client-level priority is set during client initialization with the `priority` parameter.

# Create a client with Low priority for all requests
low_priority_client = CosmosClient(url, key, priority="Low")
low_priority_database = low_priority_client.get_database_client(database_name)
low_priority_container = low_priority_database.get_container_client(container_name)

# Add some items to query
for i in range(1, 4):
    low_priority_container.upsert_item(
        dict(id="low_priority_item{}".format(i), productName="Widget", productModel="Model {}".format(i))
    )

# All requests from this client will have Low priority by default
for queried_item in low_priority_container.query_items(
    query='SELECT * FROM products p WHERE p.productName = "Widget"',
    enable_cross_partition_query=True
):
    print(json.dumps(queried_item, indent=True))

# [END client_level_priority]

# [START request_priority_precedence]
# Request-level priority takes precedence over client-level priority.
# This allows you to override the default priority for specific operations.

# Create a client with Low priority
client_with_default_priority = CosmosClient(url, key, priority="Low")
database_with_priority = client_with_default_priority.get_database_client(database_name)
container_with_priority = database_with_priority.get_container_client(container_name)

# Add items with different priority levels to the container
container_with_priority.upsert_item(
    dict(id="urgent_item1", productName="Widget", priority="High", productModel="High Priority Model")
)
container_with_priority.upsert_item(
    dict(id="normal_item1", productName="Widget", priority="Low", productModel="Low Priority Model")
)

# This query will use High priority, overriding the client's Low priority setting
for important_item in container_with_priority.query_items(
    query='SELECT * FROM products p WHERE p.priority = "High"',
    enable_cross_partition_query=True,
    priority="High"  # Request-level priority overrides client-level priority
):
    print(json.dumps(important_item, indent=True))

# This query will use the client's default Low priority
for normal_item in container_with_priority.query_items(
    query='SELECT * FROM products p WHERE p.priority = "Low"',
    enable_cross_partition_query=True
):
    print(json.dumps(normal_item, indent=True))

# [END request_priority_precedence]

# Delete items from the container.
# The Cosmos DB SQL API does not support 'DELETE' queries,
# so deletes must be done with the delete_item method
# on the container.
# [START delete_items]
for queried_item in container.query_items(
    query='SELECT * FROM products p WHERE p.productModel = "DISCONTINUED" AND p.productName="Widget"'
):
    container.delete_item(queried_item, partition_key="Widget")
# [END delete_items]

# Query items with feed range is also supported. This example
# gets all items within the feed range.
# [START query_items_feed_range]
import json

for feed_range in container.read_feed_ranges():
    for queried_item in container.query_items(
        query='SELECT * FROM c',
        feed_range=feed_range,
    ):
        print(json.dumps(queried_item, indent=True))
# [END query_items_param]

# Retrieve the properties of a database
# [START get_database_properties]
read_properties = database.read()
print(json.dumps(read_properties, indent=True))
# [END get_database_properties]

# Retrieve the properties of a container
# [START get_container_properties]
# Get properties will return a cache of two container properties: RID and the Partition Key Definition (This will not consume RUs)
properties = container._get_properties()

# Print _rid and partitionKey
print("Resource ID: ", properties.get('_rid'))
print("Partition Key: ", properties.get('partitionKey'))

# Read the container to get the latests of all the Container Properties. (This will make a backend requests and will consume RUs)
container_properties = container.read()

# Print each property one by one if they are currently in the container properties
print("indexingPolicy: ", container_properties.get("indexingPolicy"))
print("etag: ", container_properties.get('_etag'))
print("lastModified: ", container_properties.get('lastModified'))
print("defaultTtl: ", container_properties.get('defaultTtl'))
print("uniqueKeyPolicy: ", container_properties.get('uniqueKeyPolicy'))
print("conflictResolutionPolicy: ", container_properties.get('conflictResolutionPolicy'))
print("changeFeedPolicy: ", container_properties.get('changeFeedPolicy'))
print("geospatialConfig: ", container_properties.get('geospatialConfig'))

# Print remaining properties if they are in the current container properties
for key, value in container_properties.items():
    if key not in ['_rid', 'partitionKey', 'indexingPolicy', '_etag', 'lastModified', 'defaultTtl', 'uniqueKeyPolicy', 'conflictResolutionPolicy', 'changeFeedPolicy', 'geospatialConfig']:
        print(f"{key}: {value}")
# [END get_container_properties]

# Modify the properties of an existing container
# This example sets the default time to live (TTL) for items in the
# container to 3600 seconds (1 hour). An item in container is deleted
# when the TTL has elapsed since it was last edited.
# [START reset_container_properties]
# Set the TTL on the container to 3600 seconds (one hour)
database.replace_container(container, partition_key=PartitionKey(path='/productName'), default_ttl=3600)

# Display the new TTL setting for the container
container_props = database.get_container_client(container_name).read()
print("New container TTL: {}".format(json.dumps(container_props['defaultTtl'])))
# [END reset_container_properties]

# Create a user in the database.
# [START create_user]
try:
    database.create_user(dict(id="Walter Harp"))
except exceptions.CosmosResourceExistsError:
    print("A user with that ID already exists.")
except exceptions.CosmosHttpResponseError as failure:
    print("Failed to create user. Status code:{}".format(failure.status_code))
# [END create_user]

# [START delete_all_items_by_partition_key]
container_name = "products"
container = database.get_container_client(container_name)
for i in range(1, 10):
    container.upsert_item(
        dict(id="item{}".format(i), productName="Gadget", productModel="Model {}".format(i))
    )
items = container.read_all_items()
for item_dict in items:
    print(json.dumps(item_dict, indent=True))
container.delete_all_items_by_partition_key("Gadget")
print("All items in partition {} deleted.".format("Gadget"))
items = container.read_all_items()
for item_dict in items:
    print(json.dumps(item_dict, indent=True))
# [END delete_all_items_by_partition_key]

# Subpartitioning Samples Similar samples that show how to use subpartitioning
# [START create_container]
location_container_name = "locations"
try:
    container = database.create_container(
        id=location_container_name, partition_key=PartitionKey(path=["/state", "/city", "/zipcode"], kind="MultiHash")
    )
except exceptions.CosmosResourceExistsError:
    container = database.get_container_client(container_name)
# [END create_container]

# insert items in a subpartitioned container
# [START upsert_items]
for i in range(1, 10):
    container.upsert_item(
        dict(id="item{}".format(i), state="WA", city="Redmond", zipcode=98052)
    )
# [END upsert_items]

# Modify an existing item in the container
# [START update_item]
item = container.read_item("item2", partition_key=["WA", "Redmond", 98052])
item["state"] = "GA"
item["city"] = "Atlanta"
item["zipcode"] = 30363
updated_item = container.upsert_item(item)
# [END update_item]

# Query the items in a container using SQL-like syntax. This example
# gets all items whose product model hasn't been discontinued.
# [START query_items]
import json

for queried_item in container.query_items(
    query='SELECT * FROM products p WHERE p.state = "WA"',
    enable_cross_partition_query=True,
):
    print(json.dumps(queried_item, indent=True))
# [END query_items]

# [START delete_items]
for queried_item in container.query_items(
    query='SELECT * FROM products p WHERE p.state = "GA"'
):
    container.delete_item(queried_item, partition_key=["GA", "Atlanta", 30363])
# [END delete_items]

# Get the feed ranges list from container.
# [START read_feed_ranges]
feed_ranges = list(container.read_feed_ranges())
# [END read_feed_ranges]

# Get a feed range from a partition key.
# [START feed_range_from_partition_key ]
feed_range_from_pk = container.feed_range_from_partition_key(["GA", "Atlanta", 30363])
# [END feed_range_from_partition_key]

# Figure out if a feed range is a subset of another feed range.
# This example sees in which feed range from the container a feed range from a partition key is part of.
# [START is_feed_range_subset]
parent_feed_range = {}
for feed_range in feed_ranges:
    if container.is_feed_range_subset(feed_range, feed_range_from_pk):
        parent_feed_range = feed_range
        break
# [END is_feed_range_subset]

# Query a sorted list of items that were changed for one feed range
# [START query_items_change_feed]
for queried_item in container.query_items_change_feed(feed_range=feed_ranges[0]):
    print(json.dumps(queried_item, indent=True))
# [END query_items_change_feed]

# Query a sorted list of items that were changed for one feed range
# [START query_items_change_feed_from_beginning]
for queried_item in container.query_items_change_feed(feed_range=feed_ranges[0], start_time="Beginning"):
    print(json.dumps(queried_item, indent=True))
# [END query_items_change_feed_from_beginning]


# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

import unittest
import uuid

import pytest

import test_config
from azure.cosmos import PartitionKey, http_constants, exceptions
from azure.cosmos.aio import CosmosClient, DatabaseProxy


@pytest.mark.cosmosLong
class TestResourceIdsAsync(unittest.IsolatedAsyncioTestCase):
    configs = test_config.TestConfig
    host = configs.host
    masterKey = configs.masterKey
    connectionPolicy = configs.connectionPolicy
    last_headers = []
    client: CosmosClient = None
    created_database: DatabaseProxy = None

    @classmethod
    def setUpClass(cls):
        if (cls.masterKey == '[YOUR_KEY_HERE]' or
                cls.host == '[YOUR_ENDPOINT_HERE]'):
            raise Exception(
                "You must specify your Azure Cosmos account values for "
                "'masterKey' and 'host' at the top of this class to run the "
                "tests.")

    async def asyncSetUp(self):
        self.client = CosmosClient(self.host, self.masterKey)

    async def asyncTearDown(self):
        await self.client.close()

    async def test_id_unicode_validation_async(self):
        # unicode chars in Hindi for Id which translates to: "Hindi is the national language of India"
        resource_id1 = u'हिन्दी भारत की राष्ट्रीय भाषा है' + str(uuid.uuid4())  # cspell:disable-line

        # Special allowed chars for Id
        resource_id2 = "!@$%^&*()-~`'_[]{}|;:,.<>" + str(uuid.uuid4())

        # verify that databases are created with specified IDs
        created_db1 = await self.client.create_database_if_not_exists(resource_id1)
        created_db2 = await self.client.create_database_if_not_exists(resource_id2)

        assert resource_id1 == created_db1.id
        assert resource_id2 == created_db2.id

        # verify that collections are created with specified IDs
        created_collection1 = await created_db1.create_container(
            id=resource_id1,
            partition_key=PartitionKey(path='/id', kind='Hash'))
        created_collection2 = await created_db2.create_container(
            id=resource_id2,
            partition_key=PartitionKey(path='/id', kind='Hash'))

        assert resource_id1 == created_collection1.id
        assert resource_id2 == created_collection2.id

        # verify that items are created with specified IDs
        item1 = await created_collection1.upsert_item({"id": resource_id1})
        item2 = await created_collection1.upsert_item({"id": resource_id2})

        assert resource_id1 == item1.get("id")
        assert resource_id2 == item2.get("id")

        await self.client.delete_database(resource_id1)
        await self.client.delete_database(resource_id2)

    async def test_create_illegal_characters_async(self):
        database_id = str(uuid.uuid4())
        container_id = str(uuid.uuid4())
        partition_key = PartitionKey(path="/id")

        created_database = await self.client.create_database(id=database_id)
        created_container = await created_database.create_container(id=container_id, partition_key=partition_key)

        # Define errors returned by checks
        error_strings = ['Id contains illegal chars.', 'Id ends with a space or newline.']

        # Define illegal strings
        illegal_strings = [
            "ID_with_back/slash",
            "ID_\\with_forward_slashes",
            "ID_with_question?_mark",
            "ID_with_#pound",
            "ID_with_tab\t",
            "ID\r_with_return_carriage",
            "ID_with_newline\n",
            "ID_with_newline\n2",
            "ID_with_more_than_255" + "_added" * 255,
            "ID_with_trailing_spaces   "
        ]

        # test illegal resource id's for all resources
        for resource_id in illegal_strings:
            try:
                await self.client.create_database(resource_id)
                self.fail("Database create should have failed for id {}".format(resource_id))
            except ValueError as e:
                assert str(e) in error_strings
            except exceptions.CosmosHttpResponseError as e:
                assert e.status_code == http_constants.StatusCodes.BAD_REQUEST
                assert "Ensure to provide a unique non-empty string less than '255' characters." in e.message

            try:
                await created_database.create_container(id=resource_id, partition_key=partition_key)
                self.fail("Container create should have failed for id {}".format(resource_id))
            except ValueError as e:
                assert str(e) in error_strings
            except exceptions.CosmosHttpResponseError as e:
                assert e.status_code == http_constants.StatusCodes.BAD_REQUEST
                assert "Ensure to provide a unique non-empty string less than '255' characters." in e.message

            try:
                await created_container.create_item({"id": resource_id})
                self.fail("Item create should have failed for id {}".format(resource_id))
            except ValueError as e:
                assert str(e) in error_strings
            except exceptions.CosmosHttpResponseError as e:
                assert e.status_code == http_constants.StatusCodes.BAD_REQUEST
                assert "Ensure to provide a unique non-empty string less than '1024' characters." in e.message

            try:
                await created_container.upsert_item({"id": resource_id})
                self.fail("Item upsert should have failed for id {}".format(resource_id))
            except ValueError as e:
                assert str(e) in error_strings
            except exceptions.CosmosHttpResponseError as e:
                assert e.status_code == http_constants.StatusCodes.BAD_REQUEST
                assert "Ensure to provide a unique non-empty string less than '1024' characters." in e.message

        await self.client.delete_database(created_database)


if __name__ == '__main__':
    unittest.main()



# The MIT License (MIT)
# Copyright (c) 2021 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Internal class for partition health tracker for circuit breaker.
"""
import logging
import threading
import os
from typing import Any
from azure.cosmos._routing.routing_range import PartitionKeyRangeWrapper
from azure.cosmos._location_cache import EndpointOperationType
from azure.cosmos._request_object import RequestObject
from ._utils import current_time_millis
from ._constants import _Constants as Constants

MINIMUM_REQUESTS_FOR_FAILURE_RATE = 100
MAX_UNAVAILABLE_TIME_MS = 1200 * 1000 # 20 minutes in milliseconds
REFRESH_INTERVAL_MS = 60 * 1000 # 1 minute in milliseconds
INITIAL_UNAVAILABLE_TIME_MS = 60 * 1000 # 1 minute in milliseconds
# partition is unhealthy if sdk tried to recover and failed
UNHEALTHY = "unhealthy"
# partition is unhealthy tentative when it initially marked unavailable
UNHEALTHY_TENTATIVE = "unhealthy_tentative"
# unavailability info keys
UNAVAILABLE_INTERVAL = "unavailableInterval"
LAST_UNAVAILABILITY_CHECK_TIME_STAMP = "lastUnavailabilityCheckTimeStamp"
HEALTH_STATUS = "healthStatus"

class _PartitionHealthInfo(object):
    """
    This internal class keeps the health and statistics for a partition.
    """

    def __init__(self) -> None:
        self.write_failure_count: int = 0
        self.read_failure_count: int = 0
        self.write_success_count: int = 0
        self.read_success_count: int = 0
        self.read_consecutive_failure_count: int = 0
        self.write_consecutive_failure_count: int = 0
        self.unavailability_info: dict[str, Any] = {}

    def reset_failure_rate_health_stats(self) -> None:
        self.write_failure_count = 0
        self.read_failure_count = 0
        self.write_success_count = 0
        self.read_success_count = 0

    def transition_health_status(self, target_health_status: str, curr_time: int) -> None:
        if target_health_status == UNHEALTHY :
            self.unavailability_info[HEALTH_STATUS] = UNHEALTHY
            # reset the last unavailability check time stamp
            self.unavailability_info[UNAVAILABLE_INTERVAL] = \
                min(self.unavailability_info[UNAVAILABLE_INTERVAL] * 2,
                    MAX_UNAVAILABLE_TIME_MS)
            self.unavailability_info[LAST_UNAVAILABILITY_CHECK_TIME_STAMP] \
                = curr_time
        elif target_health_status == UNHEALTHY_TENTATIVE :
            self.unavailability_info = {
                LAST_UNAVAILABILITY_CHECK_TIME_STAMP: curr_time,
                UNAVAILABLE_INTERVAL: INITIAL_UNAVAILABLE_TIME_MS,
                HEALTH_STATUS: UNHEALTHY_TENTATIVE
            }

    def __str__(self) -> str:
        return (f"{self.__class__.__name__}: {self.unavailability_info}\n"
                f"write failure count: {self.write_failure_count}\n"
                f"read failure count: {self.read_failure_count}\n"
                f"write success count: {self.write_success_count}\n"
                f"read success count: {self.read_success_count}\n"
                f"write consecutive failure count: {self.write_consecutive_failure_count}\n"
                f"read consecutive failure count: {self.read_consecutive_failure_count}\n")

def _has_exceeded_failure_rate_threshold(
        successes: int,
        failures: int,
        failure_rate_threshold: int,
) -> bool:
    if successes + failures < MINIMUM_REQUESTS_FOR_FAILURE_RATE:
        return False
    failure_rate = failures / (failures + successes) * 100
    return failure_rate >= failure_rate_threshold

def _should_mark_healthy_tentative(partition_health_info: _PartitionHealthInfo, curr_time: int) -> bool:
    elapsed_time = (curr_time -
                    partition_health_info.unavailability_info[LAST_UNAVAILABILITY_CHECK_TIME_STAMP])
    current_health_status = partition_health_info.unavailability_info[HEALTH_STATUS]
    stale_partition_unavailability_check = partition_health_info.unavailability_info[UNAVAILABLE_INTERVAL]
    # check if the partition key range is still unavailable
    return ((current_health_status == UNHEALTHY and elapsed_time > stale_partition_unavailability_check)
            or (current_health_status == UNHEALTHY_TENTATIVE and elapsed_time > INITIAL_UNAVAILABLE_TIME_MS))

logger = logging.getLogger("azure.cosmos._PartitionHealthTracker")

class _PartitionHealthTracker(object):
    """
    This internal class implements the logic for tracking health thresholds for a partition.
    """

    def __init__(self) -> None:
        # partition -> regions -> health info
        self.pk_range_wrapper_to_health_info: dict[PartitionKeyRangeWrapper, dict[str, _PartitionHealthInfo]] = {}
        self.last_refresh = current_time_millis()
        self.stale_partition_lock = threading.Lock()

    def _transition_health_status_on_failure(
            self,
            pk_range_wrapper: PartitionKeyRangeWrapper,
            location: str
    ) -> None:
        logger.warning("%s has been marked as unavailable.", pk_range_wrapper)
        current_time = current_time_millis()
        if pk_range_wrapper not in self.pk_range_wrapper_to_health_info:
            # healthy -> unhealthy tentative
            partition_health_info = _PartitionHealthInfo()
            partition_health_info.transition_health_status(UNHEALTHY_TENTATIVE, current_time)
            self.pk_range_wrapper_to_health_info[pk_range_wrapper] = {
                location: partition_health_info
            }
        else:
            region_to_partition_health = self.pk_range_wrapper_to_health_info[pk_range_wrapper]
            if location in region_to_partition_health and region_to_partition_health[location].unavailability_info:
                # healthy tentative -> unhealthy
                region_to_partition_health[location].transition_health_status(UNHEALTHY, current_time)
                # if the operation type is not empty, we are in the healthy tentative state
            else:
                # healthy -> unhealthy tentative
                # if the operation type is empty, we are in the unhealthy tentative state
                partition_health_info = _PartitionHealthInfo()
                partition_health_info.transition_health_status(UNHEALTHY_TENTATIVE, current_time)
                self.pk_range_wrapper_to_health_info[pk_range_wrapper][location] = partition_health_info

    def _transition_health_status_on_success(
            self,
            pk_range_wrapper: PartitionKeyRangeWrapper,
            location: str
    ) -> None:
        if pk_range_wrapper in self.pk_range_wrapper_to_health_info:
            # healthy tentative -> healthy
            self.pk_range_wrapper_to_health_info[pk_range_wrapper][location].unavailability_info = {}

    def check_stale_partition_info(
            self,
            request: RequestObject,
            pk_range_wrapper: PartitionKeyRangeWrapper
    ) -> None:
        current_time = current_time_millis()

        if pk_range_wrapper in self.pk_range_wrapper_to_health_info:
            for location, partition_health_info in self.pk_range_wrapper_to_health_info[pk_range_wrapper].items():
                if partition_health_info.unavailability_info:
                    if _should_mark_healthy_tentative(partition_health_info, current_time):
                        # unhealthy or unhealthy tentative -> healthy tentative
                        # only one request should be used to recover
                        with self.stale_partition_lock:
                            if _should_mark_healthy_tentative(partition_health_info, current_time):
                                logger.debug("Attempting recovery for %s in %s where health info is %s.",
                                            pk_range_wrapper,
                                            location,
                                            partition_health_info)
                                # this will trigger one attempt to recover
                                partition_health_info.transition_health_status(UNHEALTHY, current_time)
                                request.healthy_tentative_location = location

        if current_time - self.last_refresh > REFRESH_INTERVAL_MS:
            # all partition stats reset every minute
            self._reset_partition_health_tracker_stats()
            self.last_refresh = current_time


    def get_unhealthy_locations(
            self,
            request: RequestObject,
            pk_range_wrapper: PartitionKeyRangeWrapper
        ) -> list[str]:
        unhealthy_locations = []
        if pk_range_wrapper in self.pk_range_wrapper_to_health_info:
            for location, partition_health_info in self.pk_range_wrapper_to_health_info[pk_range_wrapper].items():
                if (partition_health_info.unavailability_info and
                        not (request.healthy_tentative_location and request.healthy_tentative_location == location)):
                    health_status = partition_health_info.unavailability_info[HEALTH_STATUS]
                    if health_status in (UNHEALTHY_TENTATIVE, UNHEALTHY) :
                        unhealthy_locations.append(location)
        return unhealthy_locations

    def add_failure(
            self,
            pk_range_wrapper: PartitionKeyRangeWrapper,
            operation_type: str,
            location: str
    ) -> None:
        # Retrieve the failure rate threshold from the environment.
        failure_rate_threshold = int(os.environ.get(Constants.FAILURE_PERCENTAGE_TOLERATED,
                                               Constants.FAILURE_PERCENTAGE_TOLERATED_DEFAULT))

        # Ensure that the health info dictionary is properly initialized.
        if pk_range_wrapper not in self.pk_range_wrapper_to_health_info:
            self.pk_range_wrapper_to_health_info[pk_range_wrapper] = {}
        if location not in self.pk_range_wrapper_to_health_info[pk_range_wrapper]:
            self.pk_range_wrapper_to_health_info[pk_range_wrapper][location] = _PartitionHealthInfo()

        health_info = self.pk_range_wrapper_to_health_info[pk_range_wrapper][location]

        # Determine attribute names and environment variables based on the operation type.
        if operation_type == EndpointOperationType.WriteType:
            success_attr = 'write_success_count'
            failure_attr = 'write_failure_count'
            consecutive_attr = 'write_consecutive_failure_count'
            env_key = Constants.CONSECUTIVE_ERROR_COUNT_TOLERATED_FOR_WRITE
            default_consecutive_threshold = Constants.CONSECUTIVE_ERROR_COUNT_TOLERATED_FOR_WRITE_DEFAULT
        else:
            success_attr = 'read_success_count'
            failure_attr = 'read_failure_count'
            consecutive_attr = 'read_consecutive_failure_count'
            env_key = Constants.CONSECUTIVE_ERROR_COUNT_TOLERATED_FOR_READ
            default_consecutive_threshold = Constants.CONSECUTIVE_ERROR_COUNT_TOLERATED_FOR_READ_DEFAULT

        # Increment failure and consecutive failure counts.
        setattr(health_info, failure_attr, getattr(health_info, failure_attr) + 1)
        setattr(health_info, consecutive_attr, getattr(health_info, consecutive_attr) + 1)

        # Retrieve the consecutive failure threshold from the environment.
        consecutive_failure_threshold = int(os.environ.get(env_key, default_consecutive_threshold))
        # log the current stats
        logger.debug("Failure for partition %s in location %s has %s",
                    pk_range_wrapper,
                     location,
                     self.pk_range_wrapper_to_health_info[pk_range_wrapper][location])

        # Call the threshold checker with the current stats.
        self._check_thresholds(
            pk_range_wrapper,
            getattr(health_info, success_attr),
            getattr(health_info, failure_attr),
            getattr(health_info, consecutive_attr),
            location,
            failure_rate_threshold,
            consecutive_failure_threshold
        )

    def _check_thresholds(
            self,
            pk_range_wrapper: PartitionKeyRangeWrapper,
            successes: int,
            failures: int,
            consecutive_failures: int,
            location: str,
            failure_rate_threshold: int,
            consecutive_failure_threshold: int,
    ) -> None:
        # check the failure rate was not exceeded
        if _has_exceeded_failure_rate_threshold(
                successes,
                failures,
                failure_rate_threshold
        ):
            self._transition_health_status_on_failure(pk_range_wrapper, location)

        # add to consecutive failures and check that threshold was not exceeded
        if consecutive_failures >= consecutive_failure_threshold:
            self._transition_health_status_on_failure(pk_range_wrapper, location)

    def add_success(self, pk_range_wrapper: PartitionKeyRangeWrapper, operation_type: str, location: str) -> None:
        # Ensure that the health info dictionary is initialized.
        if pk_range_wrapper not in self.pk_range_wrapper_to_health_info:
            self.pk_range_wrapper_to_health_info[pk_range_wrapper] = {}
        if location not in self.pk_range_wrapper_to_health_info[pk_range_wrapper]:
            self.pk_range_wrapper_to_health_info[pk_range_wrapper][location] = _PartitionHealthInfo()

        health_info = self.pk_range_wrapper_to_health_info[pk_range_wrapper][location]

        if operation_type == EndpointOperationType.WriteType:
            health_info.write_success_count += 1
            health_info.write_consecutive_failure_count = 0
        else:
            health_info.read_success_count += 1
            health_info.read_consecutive_failure_count = 0
        self._transition_health_status_on_success(pk_range_wrapper, location)

    def _reset_partition_health_tracker_stats(self) -> None:
        for locations in self.pk_range_wrapper_to_health_info.values():
            for health_info in locations.values():
                health_info.reset_failure_rate_health_stats()



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.
import unittest
import uuid
import pytest
from azure.cosmos.aio import CosmosClient, DatabaseProxy
import test_config
from azure.cosmos import PartitionKey


@pytest.mark.cosmosSplit
class TestReadItemsPartitionSplitScenarios(unittest.IsolatedAsyncioTestCase):
    """Tests the behavior of read_items in scenarios involving partition splits."""

    created_db: DatabaseProxy = None
    client: CosmosClient = None
    host = test_config.TestConfig.host
    masterKey = test_config.TestConfig.masterKey
    configs = test_config.TestConfig
    TEST_DATABASE_ID = configs.TEST_DATABASE_ID

    async def asyncSetUp(self):
        self.client = CosmosClient(self.host, self.masterKey)
        self.database = self.client.get_database_client(self.TEST_DATABASE_ID)

    async def asyncTearDown(self):
        await self.client.close()

    async def test_read_items_with_partition_split_async(self):
        """Tests that read_items works correctly after a partition split."""
        container = await self.database.create_container(
            "read_items_split_test_async" + str(uuid.uuid4()),
            PartitionKey(path="/pk"),
            offer_throughput=400)
        # 1. Create 5 items to read
        items_to_read = []
        item_ids = []
        for i in range(5):
            doc_id = f"item_split_{i}_{uuid.uuid4()}"
            item_ids.append(doc_id)
            # Add the partition key field 'pk' to the item body
            await container.create_item({'id': doc_id, 'pk': doc_id, 'data': i})
            items_to_read.append((doc_id, doc_id))

        # 2. Initial read_items call before the split
        print("Performing initial read_items call...")
        initial_read_items = await container.read_items(items=items_to_read)
        self.assertEqual(len(initial_read_items), len(items_to_read))
        print("Initial call successful.")

        # 3. Trigger a partition split
        await test_config.TestConfig.trigger_split_async(container, 11000)

        # 4. Call read_items again after the split
        print("Performing post-split read_items call...")
        final_read_items = await container.read_items(items=items_to_read)

        # 5. Verify the results
        self.assertEqual(len(final_read_items), len(items_to_read))
        final_read_ids = {item['id'] for item in final_read_items}
        self.assertSetEqual(final_read_ids, set(item_ids))
        print("Post-split call successful.")
        await self.database.delete_container(container.id)


if __name__ == '__main__':
    unittest.main()


# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

import logging
import time
import unittest
import uuid
import test_config
import pytest
import pytest_asyncio

from azure.cosmos.aio import CosmosClient
from azure.cosmos.documents import _OperationType as OperationType
from azure.cosmos.http_constants import ResourceType
from test_excluded_locations import (TestDataType, set_test_data_type,
                                     read_item_test_data, write_item_test_data, read_and_write_item_test_data,
                                     verify_endpoint)


class MockHandler(logging.Handler):
    def __init__(self):
        super(MockHandler, self).__init__()
        self.messages = []

    def reset(self):
        self.messages = []

    def emit(self, record):
        self.messages.append(record.msg)

# Test configurations
MOCK_HANDLER = MockHandler()
CONFIG = test_config.TestConfig()
HOST = CONFIG.host
KEY = CONFIG.masterKey
DATABASE_ID = CONFIG.TEST_DATABASE_ID
CONTAINER_ID = CONFIG.TEST_MULTI_PARTITION_PREFIX_PK_CONTAINER_ID
PARTITION_KEY = CONFIG.TEST_CONTAINER_PREFIX_PARTITION_KEY
ITEM_ID = 'doc1'
PARTITION_KEY_VALUES = [f'value{i+1}' for i in range(len(PARTITION_KEY))]
PREFIX_PARTITION_KEY = [PARTITION_KEY_VALUES[0]]
PARTITION_KEY_ITEMS = dict(zip(PARTITION_KEY, PARTITION_KEY_VALUES))
TEST_ITEM = {'id': ITEM_ID}
TEST_ITEM.update(PARTITION_KEY_ITEMS)

set_test_data_type(TestDataType.ALL_TESTS)

async def create_item_with_excluded_locations_async(container, body, excluded_locations):
    if excluded_locations is None:
        await container.create_item(body=body)
    else:
        await container.create_item(body=body, excluded_locations=excluded_locations)

async def init_container_async(client):
    db = client.get_database_client(DATABASE_ID)
    container = db.get_container_client(CONTAINER_ID)
    MOCK_HANDLER.reset()

    return db, container

@pytest_asyncio.fixture(scope="class", autouse=True)
async def setup_and_teardown_async():
    print("Setup: This runs before any tests")
    logger = logging.getLogger("azure")
    logger.addHandler(MOCK_HANDLER)
    logger.setLevel(logging.DEBUG)

    test_client = CosmosClient(HOST, KEY)
    container = test_client.get_database_client(DATABASE_ID).get_container_client(CONTAINER_ID)
    await container.upsert_item(body=TEST_ITEM)
    # Waiting some time for the new items to be replicated to other regions
    time.sleep(3)
    yield
    # Code to run after tests
    print("Teardown: This runs after all tests")

@pytest.mark.cosmosCircuitBreaker
@pytest.mark.cosmosMultiRegion
@pytest.mark.asyncio
@pytest.mark.usefixtures("setup_and_teardown_async")
class TestExcludedLocationsAsync:
    @pytest.mark.parametrize('test_data', read_item_test_data())
    async def test_read_item_async(self, test_data):
        # Init test variables
        preferred_locations, client_excluded_locations, request_excluded_locations, expected_locations = test_data

        # Client setup
        async with CosmosClient(HOST, KEY,
                              preferred_locations=preferred_locations,
                              excluded_locations=client_excluded_locations,
                              multiple_write_locations=True) as client:
            db, container = await init_container_async(client)

            # API call: read_item
            if request_excluded_locations is None:
                await container.read_item(ITEM_ID, PARTITION_KEY_VALUES)
            else:
                await container.read_item(ITEM_ID, PARTITION_KEY_VALUES, excluded_locations=request_excluded_locations)

            # Verify endpoint locations
            verify_endpoint(MOCK_HANDLER.messages, client, expected_locations)

    @pytest.mark.parametrize('test_data', read_item_test_data())
    async def test_read_all_items_async(self, test_data):
        # Init test variables
        preferred_locations, client_excluded_locations, request_excluded_locations, expected_locations = test_data

        # Client setup
        async with CosmosClient(HOST, KEY,
                                preferred_locations=preferred_locations,
                                excluded_locations=client_excluded_locations,
                                multiple_write_locations=True) as client:
            db, container = await init_container_async(client)

            # API call: read_all_items
            if request_excluded_locations is None:
                all_items = [item async for item in container.read_all_items()]
            else:
                all_items = [item async for item in container.read_all_items(excluded_locations=request_excluded_locations)]

            # Verify endpoint locations
            verify_endpoint(MOCK_HANDLER.messages, client, expected_locations)

    @pytest.mark.parametrize('test_data', read_item_test_data())
    async def test_query_items_with_partition_key_async(self, test_data):
        # Init test variables
        preferred_locations, client_excluded_locations, request_excluded_locations, expected_locations = test_data

        # Client setup
        async with CosmosClient(HOST, KEY,
                                preferred_locations=preferred_locations,
                                excluded_locations=client_excluded_locations,
                                multiple_write_locations=True) as client:
            db, container = await init_container_async(client)

            # API call: query_items
            query = 'select * from c'
            if request_excluded_locations is None:
                all_items = [item async for item in container.query_items(query, partition_key=PREFIX_PARTITION_KEY)]
            else:
                all_items = [item async for item in container.query_items(query, partition_key=PREFIX_PARTITION_KEY, excluded_locations=request_excluded_locations)]

            # Verify endpoint locations
            verify_endpoint(MOCK_HANDLER.messages, client, expected_locations)

    @pytest.mark.parametrize('test_data', read_item_test_data())
    async def test_query_items_with_query_plan_async(self, test_data):
        # Init test variables
        preferred_locations, client_excluded_locations, request_excluded_locations, expected_locations = test_data

        # Client setup
        async with CosmosClient(HOST, KEY,
                                preferred_locations=preferred_locations,
                                excluded_locations=client_excluded_locations,
                                multiple_write_locations=True) as client:
            db, container = await init_container_async(client)

            # API call: query_items
            query = 'Select top 10 value count(c.id) from c'
            if request_excluded_locations is None:
                all_items = [item async for item in container.query_items(query)]
            else:
                all_items = [item async for item in container.query_items(query, excluded_locations=request_excluded_locations)]

            # Verify endpoint locations
            verify_endpoint(MOCK_HANDLER.messages, client, expected_locations)

    @pytest.mark.parametrize('test_data', read_item_test_data())
    async def test_query_items_change_feed_async(self, test_data):
        # Init test variables
        preferred_locations, client_excluded_locations, request_excluded_locations, expected_locations = test_data


        # Client setup
        async with CosmosClient(HOST, KEY,
                                preferred_locations=preferred_locations,
                                excluded_locations=client_excluded_locations,
                                multiple_write_locations=True) as client:
            db, container = await init_container_async(client)
            # API call: query_items_change_feed
            if request_excluded_locations is None:
                all_items = [item async for item in container.query_items_change_feed(start_time="Beginning", partition_key=PREFIX_PARTITION_KEY)]
            else:
                all_items = [item async for item in container.query_items_change_feed(start_time="Beginning", partition_key=PREFIX_PARTITION_KEY, excluded_locations=request_excluded_locations)]

            # Verify endpoint locations
            verify_endpoint(MOCK_HANDLER.messages, client, expected_locations)


    @pytest.mark.parametrize('test_data', read_and_write_item_test_data())
    async def test_replace_item(self, test_data):
        # Init test variables
        preferred_locations, client_excluded_locations, request_excluded_locations, expected_locations = test_data

        for multiple_write_locations in [True, False]:
            # Client setup
            async with CosmosClient(HOST, KEY,
                                    preferred_locations=preferred_locations,
                                    excluded_locations=client_excluded_locations,
                                    multiple_write_locations=multiple_write_locations) as client:
                db, container = await init_container_async(client)

                # API call: replace_item
                if request_excluded_locations is None:
                    await container.replace_item(ITEM_ID, body=TEST_ITEM)
                else:
                    await container.replace_item(ITEM_ID, body=TEST_ITEM, excluded_locations=request_excluded_locations)

                # Verify endpoint locations
                verify_endpoint(MOCK_HANDLER.messages, client, expected_locations, multiple_write_locations)

    @pytest.mark.parametrize('test_data', read_and_write_item_test_data())
    async def test_upsert_item(self, test_data):
        # Init test variables
        preferred_locations, client_excluded_locations, request_excluded_locations, expected_locations = test_data

        for multiple_write_locations in [True, False]:
            # Client setup
            async with CosmosClient(HOST, KEY,
                                    preferred_locations=preferred_locations,
                                    excluded_locations=client_excluded_locations,
                                    multiple_write_locations=multiple_write_locations) as client:
                db, container = await init_container_async(client)

                # API call: upsert_item
                body = {'id': f'doc2-{str(uuid.uuid4())}'}
                body.update(PARTITION_KEY_ITEMS)
                if request_excluded_locations is None:
                    await container.upsert_item(body=body)
                else:
                    await container.upsert_item(body=body, excluded_locations=request_excluded_locations)

                # get location from mock_handler
                verify_endpoint(MOCK_HANDLER.messages, client, expected_locations, multiple_write_locations)

    @pytest.mark.parametrize('test_data', read_and_write_item_test_data())
    async def test_create_item(self, test_data):
        # Init test variables
        preferred_locations, client_excluded_locations, request_excluded_locations, expected_locations = test_data

        for multiple_write_locations in [True, False]:
            # Client setup
            async with CosmosClient(HOST, KEY,
                                    preferred_locations=preferred_locations,
                                    excluded_locations=client_excluded_locations,
                                    multiple_write_locations=multiple_write_locations) as client:
                db, container = await init_container_async(client)

                # API call: create_item
                body = {'id': f'doc2-{str(uuid.uuid4())}'}
                body.update(PARTITION_KEY_ITEMS)
                await create_item_with_excluded_locations_async(container, body, request_excluded_locations)

                # get location from mock_handler
                verify_endpoint(MOCK_HANDLER.messages, client, expected_locations, multiple_write_locations)

    @pytest.mark.parametrize('test_data', read_and_write_item_test_data())
    async def test_patch_item_async(self, test_data):
        # Init test variables
        preferred_locations, client_excluded_locations, request_excluded_locations, expected_locations = test_data

        for multiple_write_locations in [True, False]:
            # Client setup
            async with CosmosClient(HOST, KEY,
                                    preferred_locations=preferred_locations,
                                    excluded_locations=client_excluded_locations,
                                    multiple_write_locations=multiple_write_locations) as client:
                db, container = await init_container_async(client)

                # API call: patch_item
                operations = [
                    {"op": "add", "path": "/test_data", "value": f'Data-{str(uuid.uuid4())}'},
                ]
                if request_excluded_locations is None:
                    await container.patch_item(item=ITEM_ID, partition_key=PARTITION_KEY_VALUES,
                                         patch_operations=operations)
                else:
                    await container.patch_item(item=ITEM_ID, partition_key=PARTITION_KEY_VALUES,
                                         patch_operations=operations,
                                         excluded_locations=request_excluded_locations)

                # get location from mock_handler
                verify_endpoint(MOCK_HANDLER.messages, client, expected_locations, multiple_write_locations)

    @pytest.mark.parametrize('test_data', read_and_write_item_test_data())
    async def test_execute_item_batch_async(self, test_data):
        # Init test variables
        preferred_locations, client_excluded_locations, request_excluded_locations, expected_locations = test_data

        for multiple_write_locations in [True, False]:
            # Client setup
            async with CosmosClient(HOST, KEY,
                                    preferred_locations=preferred_locations,
                                    excluded_locations=client_excluded_locations,
                                    multiple_write_locations=multiple_write_locations) as client:
                db, container = await init_container_async(client)

                # API call: execute_item_batch
                batch_operations = []
                for i in range(3):
                    body = {'id': f'doc2-{str(uuid.uuid4())}'}
                    body.update(PARTITION_KEY_ITEMS)
                    batch_operations.append(("create", (
                        body,
                    )))

                if request_excluded_locations is None:
                    await container.execute_item_batch(batch_operations=batch_operations,
                                                partition_key=PARTITION_KEY_VALUES,)
                else:
                    await container.execute_item_batch(batch_operations=batch_operations,
                                                partition_key=PARTITION_KEY_VALUES,
                                         excluded_locations=request_excluded_locations)

                # get location from mock_handler
                verify_endpoint(MOCK_HANDLER.messages, client, expected_locations, multiple_write_locations)

    @pytest.mark.parametrize('test_data', write_item_test_data())
    async def test_delete_item_async(self, test_data):
        # Init test variables
        preferred_locations, client_excluded_locations, request_excluded_locations, expected_locations = test_data

        for multiple_write_locations in [True, False]:
            # Client setup
            async with CosmosClient(HOST, KEY,
                                    preferred_locations=preferred_locations,
                                    excluded_locations=client_excluded_locations,
                                    multiple_write_locations=multiple_write_locations) as client:
                db, container = await init_container_async(client)

                # create before delete
                item_id = f'doc2-{str(uuid.uuid4())}'
                body = {'id': item_id}
                body.update(PARTITION_KEY_ITEMS)
                await create_item_with_excluded_locations_async(container, body, request_excluded_locations)
                MOCK_HANDLER.reset()

                # API call: delete_item
                if request_excluded_locations is None:
                    await container.delete_item(item_id, PARTITION_KEY_VALUES)
                else:
                    await container.delete_item(item_id, PARTITION_KEY_VALUES, excluded_locations=request_excluded_locations)

                # Verify endpoint locations
                verify_endpoint(MOCK_HANDLER.messages, client, expected_locations, multiple_write_locations,
                                operation_type=OperationType.Delete, resource_type=ResourceType.Document)

if __name__ == "__main__":
    unittest.main()



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

"""Internal class for aggregation queries implementation in the Azure Cosmos
database service.
"""
from abc import abstractmethod, ABCMeta
from azure.cosmos._execution_context.document_producer import _OrderByHelper


class _Aggregator(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def aggregate(self, other):
        pass

    @abstractmethod
    def get_result(self):
        pass


class _AverageAggregator(_Aggregator):
    def __init__(self):
        self.sum = None
        self.count = None

    def aggregate(self, other):
        if other is None or not "sum" in other:
            return
        if self.sum is None:
            self.sum = 0.0
            self.count = 0
        self.sum += other["sum"]
        self.count += other["count"]

    def get_result(self):
        if self.sum is None or self.count is None or self.count <= 0:
            return None
        return self.sum / self.count


class _CountAggregator(_Aggregator):
    def __init__(self):
        self.count = 0

    def aggregate(self, other):
        self.count += other

    def get_result(self):
        return self.count


class _MinAggregator(_Aggregator):
    def __init__(self):
        self.value = None

    def aggregate(self, other):
        if self.value is None:
            self.value = other
        else:
            if _OrderByHelper.compare({"item": other}, {"item": self.value}) < 0:
                self.value = other

    def get_result(self):
        return self.value


class _MaxAggregator(_Aggregator):
    def __init__(self):
        self.value = None

    def aggregate(self, other):
        if self.value is None:
            self.value = other
        else:
            if _OrderByHelper.compare({"item": other}, {"item": self.value}) > 0:
                self.value = other

    def get_result(self):
        return self.value


class _SumAggregator(_Aggregator):
    def __init__(self):
        self.sum = None

    def aggregate(self, other):
        if other is None:
            return
        if self.sum is None:
            self.sum = other
        else:
            self.sum += other

    def get_result(self):
        return self.sum



# The MIT License (MIT)
# Copyright (c) 2014 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Create partition keys in the Azure Cosmos DB SQL API service.
"""
from io import BytesIO
import binascii
import struct
from typing import Any, IO, Sequence, Type, Union, cast, overload
from typing_extensions import Literal

from ._cosmos_integers import _UInt32, _UInt64, _UInt128
from ._cosmos_murmurhash3 import murmurhash3_128 as _murmurhash3_128, murmurhash3_32 as _murmurhash3_32
from ._routing.routing_range import Range as _Range


_MaximumExclusiveEffectivePartitionKey = 0xFF
_MinimumInclusiveEffectivePartitionKey = 0x00
_MaxStringChars = 100
_MaxStringBytesToAppend = 100
_MaxPartitionKeyBinarySize = \
    (1  # type marker
     + 9  # hash value
     + 1  # type marker
     + _MaxStringBytesToAppend
     + 1  # trailing zero
     ) * 3


class _PartitionKeyComponentType:
    Undefined = 0x0
    Null = 0x1
    PFalse = 0x2
    PTrue = 0x3
    MinNumber = 0x4
    Number = 0x5
    MaxNumber = 0x6
    MinString = 0x7
    String = 0x8
    MaxString = 0x9
    Int64 = 0xA
    Int32 = 0xB
    Int16 = 0xC
    Int8 = 0xD
    Uint64 = 0xE
    Uint32 = 0xF
    Uint16 = 0x10
    Uint8 = 0x11
    Binary = 0x12
    Guid = 0x13
    Float = 0x14
    Infinity = 0xFF

class _PartitionKeyKind:
    HASH: str = "Hash"
    MULTI_HASH: str = "MultiHash"

class _PartitionKeyVersion:
    V1: int = 1
    V2: int = 2

class NonePartitionKeyValue:
    """Represents partition key missing from the document.
    """

class NullPartitionKeyValue:
    """Represents null value for a partition key.
    """

class _Empty:
    """Represents empty value for partitionKey when it's missing in an item belonging
    to a migrated container.
    """


class _Undefined:
    """Represents undefined value for partitionKey when it's missing in an item belonging
    to a multi-partition container.
    """


class _Infinity:
    """Represents infinity value for partitionKey."""

_SingularPartitionKeyType = Union[None, bool, float, int, str, Type[NonePartitionKeyValue], Type[NullPartitionKeyValue], _Empty, _Undefined] # pylint: disable=line-too-long
_SequentialPartitionKeyType = Sequence[_SingularPartitionKeyType]
PartitionKeyType = Union[_SingularPartitionKeyType, _SequentialPartitionKeyType]

class PartitionKey(dict):
    """Key used to partition a container into logical partitions.

    See https://learn.microsoft.com/azure/cosmos-db/partitioning-overview#choose-partitionkey
    for information on how to choose partition keys.

    This constructor supports multiple overloads:

    1. **Single Partition Key**:

       **Parameters**:
        - `path` (str): The path of the partition key.
        - `kind` (Literal["Hash"], optional): The kind of partition key. Defaults to "Hash".
        - `version` (int, optional): The version of the partition key. Defaults to 2.

       **Example**:
         >>> pk = PartitionKey(path="/id")

    2. **Hierarchical Partition Key**:

       **Parameters**:
        - `path` (list[str]): A list of paths representing the partition key, supports up to three hierarchical levels.
        - `kind` (Literal["MultiHash"], optional): The kind of partition key. Defaults to "MultiHash".
        - `version` (int, optional): The version of the partition key. Defaults to 2.

       **Example**:
         >>> pk = PartitionKey(path=["/id", "/category"], kind="MultiHash")

    :ivar str path: The path(s) of the partition key.
    :ivar str kind: The kind of partition key ("Hash" or "MultiHash") (default: "Hash").
    :ivar int version: The version of the partition key (default: 2).
    """

    @overload
    def __init__(self, path: list[str], *, kind: Literal["MultiHash"] = "MultiHash",
                 version: int = _PartitionKeyVersion.V2
    ) -> None:
        ...

    @overload
    def __init__(self, path: str, *, kind: Literal["Hash"] = "Hash",
                 version:int = _PartitionKeyVersion.V2
    ) -> None:
        ...

    def __init__(self, *args, **kwargs):
        path = args[0] if args else kwargs['path']
        kind = args[1] if len(args) > 1 else kwargs.get('kind', _PartitionKeyKind.HASH if isinstance(path, str)
        else _PartitionKeyKind.MULTI_HASH)
        version = args[2] if len(args) > 2 else kwargs.get('version', _PartitionKeyVersion.V2)
        super().__init__(paths=[path] if isinstance(path, str) else path, kind=kind, version=version)

    def __repr__(self) -> str:
        return "<PartitionKey [{}]>".format(self.path)[:1024]

    @property
    def kind(self) -> Literal["MultiHash", "Hash"]:
        return self["kind"]

    @kind.setter
    def kind(self, value: Literal["MultiHash", "Hash"]) -> None:
        self["kind"] = value

    @property
    def path(self) -> str:
        if self.kind == _PartitionKeyKind.MULTI_HASH:
            return ''.join(self["paths"])
        return self["paths"][0]

    @path.setter
    def path(self, value: Union[str, list[str]]) -> None:
        if isinstance(value, str):
            self["paths"] = [value]
        else:
            self["paths"] = value

    @property
    def version(self) -> int:
        return self["version"]

    @version.setter
    def version(self, value: int) -> None:
        self["version"] = value

    def _get_epk_range_for_prefix_partition_key(
        self,
        pk_value: _SequentialPartitionKeyType
    ) -> _Range:
        if self.kind != _PartitionKeyKind.MULTI_HASH:
            raise ValueError(
                "Effective Partition Key Range for Prefix Partition Keys is only supported for Hierarchical Partition Keys.")  # pylint: disable=line-too-long
        len_pk_value = len(pk_value)
        len_paths = len(self["paths"])
        if len_pk_value >= len_paths:
            raise ValueError(
                f"{len_pk_value} partition key components provided. Expected less than {len_paths} " +
                "components (number of container partition key definition components)."
            )
        # Prefix Partitions always have exclusive max
        min_epk = self._get_effective_partition_key_string(pk_value)
        if min_epk == _MinimumInclusiveEffectivePartitionKey:
            min_epk = ""
            return _Range(min_epk, min_epk, True, False)

        if min_epk == _MaximumExclusiveEffectivePartitionKey:
            return _Range("FF", "FF", True, False)

        max_epk = str(min_epk) + "FF"
        return _Range(min_epk, max_epk, True, False)

    def _get_epk_range_for_partition_key(
            self,
            pk_value: PartitionKeyType
    ) -> _Range:
        if self._is_prefix_partition_key(pk_value):
            return self._get_epk_range_for_prefix_partition_key(
                cast(_SequentialPartitionKeyType, pk_value))

        # else return point range
        if isinstance(pk_value, (list, tuple)) or (isinstance(pk_value, Sequence) and not isinstance(pk_value, str)):
            effective_partition_key_string = self._get_effective_partition_key_string(pk_value)
        else:
            effective_partition_key_string =\
                self._get_effective_partition_key_string([pk_value])
        return _Range(effective_partition_key_string, effective_partition_key_string, True, True)

    @staticmethod
    def _truncate_for_v1_hashing(
            value: _SingularPartitionKeyType
    ) -> _SingularPartitionKeyType:
        if isinstance(value, str):
            return value[:100]
        return value

    @staticmethod
    def _get_effective_partition_key_for_hash_partitioning(
            pk_value: Union[str, _SequentialPartitionKeyType]
    ) -> str:
        truncated_components = []
        # In Python, Strings are sequences, so we make sure we instead hash the entire string instead of each character
        if isinstance(pk_value, str):
            truncated_components.append(PartitionKey._truncate_for_v1_hashing(pk_value))
        else:
            truncated_components = [PartitionKey._truncate_for_v1_hashing(v) for v in pk_value]
        with BytesIO() as ms:
            for component in truncated_components:
                if isinstance(component, int) and not isinstance(component, bool):
                    component = float(int(_UInt32(component)))
                PartitionKey._write_for_hashing(component, ms)

            ms_bytes: bytes = ms.getvalue()
            # We use Our own MurmurHash3 implementation to match the behavior of other SDKs
            # We put into a Cosmos Integer of Unsigned 32-bit Integer, to match the behavior of other SDKs
            hash_as_int: _UInt32 = _murmurhash3_32(bytearray(ms_bytes), 0)
            hash_value = float(int(hash_as_int))

        partition_key_components = [hash_value] + truncated_components
        return _to_hex_encoded_binary_string_v1(partition_key_components)

    @staticmethod
    def _get_hashed_partition_key_string(
            pk_value: _SequentialPartitionKeyType,
            kind: str,
            version: int = _PartitionKeyVersion.V2,
    ) -> Union[int, str]:
        if not pk_value:
            return _MinimumInclusiveEffectivePartitionKey

        if kind == _PartitionKeyKind.HASH:
            if version == _PartitionKeyVersion.V1:
                return PartitionKey._get_effective_partition_key_for_hash_partitioning(pk_value)
            if version == _PartitionKeyVersion.V2:
                return PartitionKey._get_effective_partition_key_for_hash_partitioning_v2(pk_value)
        elif kind == _PartitionKeyKind.MULTI_HASH:
            return PartitionKey._get_effective_partition_key_for_multi_hash_partitioning_v2(pk_value)
        return _to_hex_encoded_binary_string(pk_value)

    def _get_effective_partition_key_string(
        self,
        pk_value: _SequentialPartitionKeyType
    ) -> Union[int, str]:
        if isinstance(self, _Infinity):
            return _MaximumExclusiveEffectivePartitionKey

        return PartitionKey._get_hashed_partition_key_string(pk_value=pk_value, kind=self.kind, version=self.version)

    @staticmethod
    def _write_for_hashing(
            value: _SingularPartitionKeyType,
            writer: IO[bytes]
    ) -> None:
        PartitionKey._write_for_hashing_core(value, bytes([0]), writer)

    @staticmethod
    def _write_for_hashing_v2(
        value: _SingularPartitionKeyType,
        writer: IO[bytes]
    ) -> None:
        PartitionKey._write_for_hashing_core(value, bytes([0xFF]), writer)

    @staticmethod
    def _write_for_hashing_core(
        value: _SingularPartitionKeyType,
        string_suffix: bytes,
        writer: IO[bytes]
    ) -> None:
        if value is True:
            writer.write(bytes([_PartitionKeyComponentType.PTrue]))
        elif value is False:
            writer.write(bytes([_PartitionKeyComponentType.PFalse]))
        elif value is None or value == {} or value == NonePartitionKeyValue:
            writer.write(bytes([_PartitionKeyComponentType.Null]))
        elif isinstance(value, int):
            writer.write(bytes([_PartitionKeyComponentType.Number]))
            # Cast to Float to ensure correct packing
            writer.write(struct.pack('<d', float(value)))
        elif isinstance(value, float):
            writer.write(bytes([_PartitionKeyComponentType.Number]))
            writer.write(struct.pack('<d', value))
        elif isinstance(value, str):
            writer.write(bytes([_PartitionKeyComponentType.String]))
            writer.write(value.encode('utf-8'))
            writer.write(string_suffix)
        elif isinstance(value, _Undefined):
            writer.write(bytes([_PartitionKeyComponentType.Undefined]))

    @staticmethod
    def _get_effective_partition_key_for_hash_partitioning_v2(
        pk_value: _SequentialPartitionKeyType
    ) -> str:
        with BytesIO() as ms:
            for component in pk_value:
                PartitionKey._write_for_hashing_v2(component, ms)

            ms_bytes = ms.getvalue()
            hash128 = _murmurhash3_128(bytearray(ms_bytes), _UInt128(0, 0))
            hash_bytes = _UInt128.to_byte_array(hash128)
            hash_bytes.reverse()

            # Reset 2 most significant bits, as max exclusive value is 'FF'.
            # Plus one more just in case.
            hash_bytes[0] &= 0x3F

        return ''.join('{:02X}'.format(x) for x in hash_bytes)

    @staticmethod
    def _get_effective_partition_key_for_multi_hash_partitioning_v2(
        pk_value: _SequentialPartitionKeyType
    ) -> str:
        sb = []
        for value in pk_value:
            ms = BytesIO()
            binary_writer = ms  # In Python, you can write bytes directly to a BytesIO object

            # Assuming paths[i] is the correct object to call write_for_hashing_v2 on
            PartitionKey._write_for_hashing_v2(value, binary_writer)

            ms_bytes = ms.getvalue()
            hash128 = _murmurhash3_128(bytearray(ms_bytes), _UInt128(0, 0))
            hash_v_bytes = hash128.to_byte_array()
            hash_v = list(reversed(hash_v_bytes))

            # Reset 2 most significant bits, as max exclusive value is 'FF'.
            # Plus one more just in case.
            hash_v[0] &= 0x3F
            sb.append(_to_hex(bytearray(hash_v), 0, len(hash_v)))

        return ''.join(sb).upper()

    def _is_prefix_partition_key(
            self,
            partition_key: PartitionKeyType) -> bool:  # pylint: disable=line-too-long
        if self.kind != _PartitionKeyKind.MULTI_HASH:
            return False
        ret = ((isinstance(partition_key, Sequence) and
                not isinstance(partition_key, str)) and len(self['paths']) != len(partition_key))
        return ret


def _return_undefined_or_empty_partition_key(is_system_key: bool) -> Union[_Empty, _Undefined]:
    if is_system_key:
        return _Empty()
    return _Undefined()


def _to_hex(bytes_object: bytearray, start: int, length: int) -> str:
    return binascii.hexlify(bytes_object[start:start + length]).decode()


def _to_hex_encoded_binary_string(components: Sequence[object]) -> str:
    buffer_bytes = bytearray(_MaxPartitionKeyBinarySize)
    ms = BytesIO(buffer_bytes)

    for component in components:
        if isinstance(component, (bool, int, float, str, _Infinity, _Undefined)):
            component = cast(_SingularPartitionKeyType, component)
            _write_for_binary_encoding(component, ms)
        else:
            raise TypeError(f"Unexpected type for PK component: {type(component)}")

    return _to_hex(buffer_bytes[:ms.tell()], 0, ms.tell())

def _to_hex_encoded_binary_string_v1(components: Sequence[object]) -> str:
    ms = BytesIO()
    for component in components:
        if (isinstance(component, (bool, int, float, str, _Infinity, _Undefined, type))
                or component is None):
            component = cast(_SingularPartitionKeyType, component)
            _write_for_binary_encoding_v1(component, ms)
        else:
            raise TypeError(f"Unexpected type for PK component: {type(component)}")

    return _to_hex(bytearray(ms.getvalue()), 0, ms.tell())

def _write_for_binary_encoding_v1(
    value: _SingularPartitionKeyType,
    binary_writer: IO[bytes]
) -> None:
    if isinstance(value, bool):
        binary_writer.write(bytes([(_PartitionKeyComponentType.PTrue if value else _PartitionKeyComponentType.PFalse)]))

    elif isinstance(value, _Infinity):
        binary_writer.write(bytes([_PartitionKeyComponentType.Infinity]))

    elif isinstance(value, (int, float)):  # Assuming number value is int or float
        binary_writer.write(bytes([_PartitionKeyComponentType.Number]))
        # For V1 Hashing we need to encode the value as a UInt64 From a Float regardless if it was an int or float
        if isinstance(value, float):
            payload = _UInt64(_UInt64.encode_double_as_uint64(value))
        else:
            payload = _UInt64(_UInt64.encode_double_as_uint64(float(value)))

        # Encode first chunk with 8-bits of payload
        binary_writer.write(bytes([int((payload >> (64 - 8)))]))
        payload <<= 8

        # Encode remaining chunks with 7 bits of payload followed by single "1" bit each.
        byte_to_write = 0
        first_iteration = True
        while payload != 0:
            if not first_iteration:
                binary_writer.write(bytes([byte_to_write]))
            else:
                first_iteration = False

            byte_to_write = int((payload >> (64 - 8)) | int(0x01))
            payload <<= 7

        # Except for last chunk that ends with "0" bit.
        binary_writer.write(bytes([(byte_to_write & 0xFE)]))

    elif isinstance(value, str):
        binary_writer.write(bytes([_PartitionKeyComponentType.String]))
        utf8_value = value.encode('utf-8')
        short_string = len(utf8_value) <= _MaxStringBytesToAppend

        for index in range(short_string and len(utf8_value) or _MaxStringBytesToAppend + 1):
            char_byte = utf8_value[index]
            char_byte += 1
            binary_writer.write(bytes([char_byte]))

        if short_string:
            binary_writer.write(bytes([0x00]))

    elif isinstance(value, _Undefined):
        binary_writer.write(bytes([_PartitionKeyComponentType.Undefined]))

def _write_for_binary_encoding(
    value: _SingularPartitionKeyType,
    binary_writer: IO[bytes]
) -> None:
    if isinstance(value, bool):
        binary_writer.write(bytes([(_PartitionKeyComponentType.PTrue if value else _PartitionKeyComponentType.PFalse)]))

    elif isinstance(value, _Infinity):
        binary_writer.write(bytes([_PartitionKeyComponentType.Infinity]))

    elif isinstance(value, (int, float)):  # Assuming number value is int or float
        binary_writer.write(bytes([_PartitionKeyComponentType.Number]))
        payload = _UInt64.encode_double_as_uint64(value)  # Function to be defined elsewhere

        # Encode first chunk with 8-bits of payload
        binary_writer.write(bytes([(payload >> (64 - 8))]))
        payload <<= 8

        # Encode remaining chunks with 7 bits of payload followed by single "1" bit each.
        byte_to_write = 0
        first_iteration = True
        while payload != 0:
            if not first_iteration:
                binary_writer.write(bytes([byte_to_write]))
            else:
                first_iteration = False

            byte_to_write = (payload >> (64 - 8)) | 0x01
            payload <<= 7

        # Except for last chunk that ends with "0" bit.
        binary_writer.write(bytes([(byte_to_write & 0xFE)]))

    elif isinstance(value, str):
        binary_writer.write(bytes([_PartitionKeyComponentType.String]))
        utf8_value = value.encode('utf-8')
        short_string = len(utf8_value) <= _MaxStringBytesToAppend

        for index in range(short_string and len(utf8_value) or _MaxStringBytesToAppend + 1):
            char_byte = utf8_value[index]
            if char_byte < 0xFF:
                char_byte += 1
            binary_writer.write(bytes([char_byte]))

        if short_string:
            binary_writer.write(bytes([0x00]))

    elif isinstance(value, _Undefined):
        binary_writer.write(bytes([_PartitionKeyComponentType.Undefined]))

def _get_partition_key_from_partition_key_definition(
    partition_key_definition: Union[dict[str, Any], "PartitionKey"]
) -> "PartitionKey":
    """Internal method to create a PartitionKey instance from a dictionary or PartitionKey object.

    :param partition_key_definition: A dictionary or PartitionKey object containing the partition key definition.
    :type partition_key_definition: Union[dict[str, Any], PartitionKey]
    :return: A PartitionKey instance created from the provided definition.
    :rtype: PartitionKey
    """
    path = partition_key_definition.get("paths", "")
    kind = partition_key_definition.get("kind", "Hash")
    version: int = partition_key_definition.get("version", 1)  # Default to version 1 if not provided
    return PartitionKey(path=path, kind=kind, version=version)

def _build_partition_key_from_properties(container_properties: dict[str, Any]) -> PartitionKey:
    partition_key_definition = container_properties["partitionKey"]
    return _get_partition_key_from_partition_key_definition(partition_key_definition)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
from azure.cosmos.aio import CosmosClient
import azure.cosmos.exceptions as exceptions
from azure.cosmos.partition_key import PartitionKey
from azure.cosmos import ThroughputProperties

import asyncio
import config

# ----------------------------------------------------------------------------------------------------------
# Prerequisites -
#
# 1. An Azure Cosmos account -
#    https://azure.microsoft.com/documentation/articles/documentdb-create-account/
#
# 2. Microsoft Azure Cosmos PyPi package -
#    https://pypi.python.org/pypi/azure-cosmos/
# ----------------------------------------------------------------------------------------------------------
# Sample - demonstrates the basic CRUD operations on a Container resource for Azure Cosmos
#
# 1. Query for Container
#
# 2. Create Container
#    2.1 - Basic Create
#    2.2 - Create container with custom IndexPolicy
#    2.3 - Create container with provisioned throughput set
#    2.4 - Create container with unique key
#    2.5 - Create Container with partition key V2
#    2.6 - Create Container with partition key V1
#    2.7 - Create Container with analytical store enabled
#
# 3. Manage Container Provisioned Throughput
#    3.1 - Get Container provisioned throughput (RU/s)
#    3.2 - Change provisioned throughput (RU/s)
#
# 4. Get a Container by its Id property
#
# 5. List all Container resources in a Database
#
# 6. Delete Container
# ----------------------------------------------------------------------------------------------------------
# Note -
#
# Running this sample will create (and delete) multiple Containers on your account.
# Each time a Container is created the account will be billed for 1 hour of usage based on
# the provisioned throughput (RU/s) of that account.
# ----------------------------------------------------------------------------------------------------------

HOST = config.settings['host']
MASTER_KEY = config.settings['master_key']
DATABASE_ID = config.settings['database_id']
CONTAINER_ID = config.settings['container_id']


async def find_container(db, id):
    print('1. Query for Container')

    # Because the asynchronous client returns an asynchronous iterator object for methods that use
    # return several containers using queries, we do not need to await the function. However, attempting
    # to cast this object into a list directly will throw an error; instead, iterate over the containers
    # to populate your list using an async for loop like shown here or in the list_containers() method
    query_containers_response = db.query_containers(
        query="SELECT * FROM r WHERE r.id=@id",
        parameters=[
            {"name": "@id", "value": id}
        ]
    )
    containers = [container async for container in query_containers_response]

    if len(containers) > 0:
        print('Container with id \'{0}\' was found'.format(id))
    else:
        print('No container with id \'{0}\' was found'.format(id))

    # Alternatively, you can directly iterate over the asynchronous iterator without building a separate
    # list if you don't need the ordering or indexing capabilities
    async for container in query_containers_response:
        print(container['id'])


async def create_container(db, id):
    """ Execute basic container creation.
    This will create containers with 400 RUs with different indexing, partitioning, and storage options """

    partition_key = PartitionKey(path='/id', kind='Hash')
    print("\n2.1 Create Container - Basic")

    try:
        await db.create_container(id=id, partition_key=partition_key)
        print('Container with id \'{0}\' created'.format(id))

    except exceptions.CosmosResourceExistsError:
        print('A container with id \'{0}\' already exists'.format(id))

    # Alternatively, you can also use the create_container_if_not_exists method to avoid using a try catch
    # This method attempts to read the container first, and based on the result either creates or returns
    # the existing container. Due to the additional overhead from attempting a read, it is recommended
    # to use the create_container() method if you know the container doesn't already exist.
    await db.create_container_if_not_exists(id=id, partition_key=partition_key)

    print("\n2.2 Create Container - With custom index policy")

    try:
        coll = {
            "id": id+"_container_custom_index_policy",
            "indexingPolicy": {
                "automatic": False
            }
        }

        container = await db.create_container(
            id=coll['id'],
            partition_key=partition_key,
            indexing_policy=coll['indexingPolicy']
        )

        properties = await container.read()
        print('Container with id \'{0}\' created'.format(container.id))
        print('IndexPolicy Mode - \'{0}\''.format(properties['indexingPolicy']['indexingMode']))
        print('IndexPolicy Automatic - \'{0}\''.format(properties['indexingPolicy']['automatic']))

    except exceptions.CosmosResourceExistsError:
        print('A container with id \'{0}\' already exists'.format(coll['id']))

    print("\n2.3 Create Container - With custom provisioned throughput")

    try:
        container = await db.create_container(
            id=id + "_container_custom_throughput",
            partition_key=partition_key,
            offer_throughput=400
        )
        print('Container with id \'{0}\' created'.format(container.id))

    except exceptions.CosmosResourceExistsError:
        print('A container with id \'{0}\' already exists'.format(coll['id']))

    print("\n2.4 Create Container - With Unique keys")

    try:
        container = await db.create_container(
            id=id + "_container_unique_keys",
            partition_key=partition_key,
            unique_key_policy={'uniqueKeys': [{'paths': ['/field1/field2', '/field3']}]}
        )
        properties = await container.read()
        unique_key_paths = properties['uniqueKeyPolicy']['uniqueKeys'][0]['paths']
        print('Container with id \'{0}\' created'.format(container.id))
        print('Unique Key Paths - \'{0}\', \'{1}\''.format(unique_key_paths[0], unique_key_paths[1]))

    except exceptions.CosmosResourceExistsError:
        print('A container with id \'container_unique_keys\' already exists')

    print("\n2.5 Create Container - With Partition key V2 (Default)")

    try:
        container = await db.create_container(
            id=id + "_container_partition_key_v2",
            partition_key=PartitionKey(path='/id', kind='Hash')
        )
        properties = await container.read()
        print('Container with id \'{0}\' created'.format(container.id))
        print('Partition Key - \'{0}\''.format(properties['partitionKey']))

    except exceptions.CosmosResourceExistsError:
        print('A container with id \'container_partition_key_v2\' already exists')

    print("\n2.6 Create Container - With Partition key V1")

    try:
        container = await db.create_container(
            id=id + "_container_partition_key_v1",
            partition_key=PartitionKey(path='/id', kind='Hash', version=1)
        )
        properties = await container.read()
        print('Container with id \'{0}\' created'.format(container.id))
        print('Partition Key - \'{0}\''.format(properties['partitionKey']))

    except exceptions.CosmosResourceExistsError:
        print('A container with id \'container_partition_key_v1\' already exists')
    except Exception:
        print("Skipping this step, account does not have Synapse Link activated")

    print("\n2.7 Create Container - With analytical store enabled")

    if 'localhost:8081' in HOST:
        print("Skipping step since emulator does not support this yet")
    else:
        try:
            container = await db.create_container(
                id=id + "_container_analytical_store",
                partition_key=PartitionKey(path='/id', kind='Hash'), analytical_storage_ttl=-1

            )
            properties = await container.read()
            print('Container with id \'{0}\' created'.format(container.id))
            print('Partition Key - \'{0}\''.format(properties['partitionKey']))

        except exceptions.CosmosResourceExistsError:
            print('A container with id \'_container_analytical_store\' already exists')
        except Exception:
            print(
                'Creating container with analytical storage can only happen in synapse link activated accounts, skipping step')

    print("\n2.8 Create Container - With autoscale settings")

    try:
        container = await db.create_container(
            id=id + "_container_auto_scale_settings",
            partition_key=partition_key,
            offer_throughput=ThroughputProperties(auto_scale_max_throughput=5000, auto_scale_increment_percent=0)
        )
        print('Container with id \'{0}\' created'.format(container.id))

    except exceptions.CosmosResourceExistsError:
        print('A container with id \'{0}\' already exists'.format(coll['id']))


async def manage_provisioned_throughput(db, id):
    print("\n3.1 Get Container provisioned throughput (RU/s)")

    # A Container's Provisioned Throughput determines the performance throughput of a container.
    # A Container is loosely coupled to Offer through the Offer's offerResourceId
    # Offer.offerResourceId == Container._rid
    # Offer.resource == Container._self

    try:
        # read the container, so we can get its _self
        container = db.get_container_client(id)

        # now use its _self to query for throughput offers
        offer = await container.get_throughput()

        print('Found Offer \'{0}\' for Container \'{1}\' and its throughput is \'{2}\''.format(offer.properties['id'],
                                                                                               container.id,
                                                                                               offer.properties[
                                                                                                   'content'][
                                                                                                   'offerThroughput']))

    except exceptions.CosmosResourceExistsError:
        print('A container with id \'{0}\' does not exist'.format(id))

    print("\n3.2 Change Provisioned Throughput of Container")

    # The Provisioned Throughput of a container controls the throughput allocated to the Container

    # The following code shows how you can change Container's throughput
    offer = await container.replace_throughput(offer.offer_throughput + 100)
    print(
        'Replaced Offer. Provisioned Throughput is now \'{0}\''.format(offer.properties['content']['offerThroughput']))


async def read_container(db, id):
    print("\n4. Get a Container by id")

    try:
        container = db.get_container_client(id)
        await container.read()
        print('Container with id \'{0}\' was found, it\'s link is {1}'.format(container.id, container.container_link))

    except exceptions.CosmosResourceNotFoundError:
        print('A container with id \'{0}\' does not exist'.format(id))


async def list_containers(db):
    print("\n5. List all Container in a Database")

    print('Containers:')

    # Because the asynchronous client returns an asynchronous iterator object for methods that use
    # return several containers using queries, we do not need to await the function. However, attempting
    # to cast this object into a list directly will throw an error; instead, iterate over the containers
    # to populate your list using an async for loop like shown here or in the find_container() method
    container_list = db.list_containers()
    containers = [container async for container in container_list]

    if len(containers) == 0:
        return

    for container in containers:
        print(container['id'])

    # Alternatively, you can directly iterate over the asynchronous iterator without building a separate
    # list if you don't need the ordering or indexing capabilities
    async for container in container_list:
        print(container['id'])


async def delete_container(db, id):
    print("\n6. Delete Container")

    try:
        await db.delete_container(id)
        print('Container with id \'{0}\' was deleted'.format(id))

    except exceptions.CosmosResourceNotFoundError:
        print('A container with id \'{0}\' does not exist'.format(id))


async def run_sample():
    async with CosmosClient(HOST, {'masterKey': MASTER_KEY}) as client:
        try:
            db = await client.create_database_if_not_exists(id=DATABASE_ID)

            # query for a container
            await find_container(db, CONTAINER_ID)

            # create a container
            await create_container(db, CONTAINER_ID)

            # get & change Provisioned Throughput of container
            await manage_provisioned_throughput(db, CONTAINER_ID)

            # get a container using its id
            await read_container(db, CONTAINER_ID)

            # list all container on an account
            await list_containers(db)

            # delete container by id
            await delete_container(db, CONTAINER_ID)

            # cleanup database after sample
            try:
                await client.delete_database(db)

            except exceptions.CosmosResourceNotFoundError:
                pass

        except exceptions.CosmosHttpResponseError as e:
            print('\nrun_sample has caught an error. {0}'.format(e.message))

        finally:
            print("\nrun_sample done")

if __name__ == '__main__':
    asyncio.run(run_sample())



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import azure.cosmos.cosmos_client as cosmos_client
from azure.cosmos.partition_key import PartitionKey
import config

# ----------------------------------------------------------------------------------------------------------
# Prerequisites -
#
# 1. An Azure Cosmos account -
#    https://azure.microsoft.com/documentation/articles/documentdb-create-account/
#
# 2. Microsoft Azure Cosmos PyPi package -
#    https://pypi.python.org/pypi/azure-cosmos/
# ----------------------------------------------------------------------------------------------------------
# Sample - demonstrates how to configure custom user agents for Azure Cosmos DB clients
#
# User agents help identify which application or process is making requests to Cosmos DB.
# Custom user agents are particularly useful for:
#   - Identifying different applications using the same Cosmos DB account
#   - Distinguishing between different versions of your application
#   - Tracking requests from specific microservices or processes
#   - Debugging and monitoring application behavior in diagnostics
#   - Analyzing usage patterns across different client applications
#
# The user agent string appears in Cosmos DB diagnostics, logs, and monitoring tools,
# making it easier to pinpoint which client instance is making specific requests.
# ----------------------------------------------------------------------------------------------------------
# Note -
#
# Running this sample will create (and delete) multiple Containers on your account.
# Each time a Container is created the account will be billed for 1 hour of usage based on
# the provisioned throughput (RU/s) of that account.
# ----------------------------------------------------------------------------------------------------------

HOST = config.settings['host']
MASTER_KEY = config.settings['master_key']
DATABASE_ID = config.settings['database_id']
CONTAINER_ID = config.settings['container_id']


def create_client_with_default_user_agent():
    """
    Create a Cosmos DB client with the default user agent.
    
    The default user agent includes information about the SDK version and Python version.
    Example: "azsdk-python-cosmos/4.5.0 Python/3.11.0 (Windows-10-10.0.22621-SP0)"
    """
    print('\n1. Creating client with default user agent')
    
    client = cosmos_client.CosmosClient(HOST, {'masterKey': MASTER_KEY})
    
    print('Client created with default user agent')
    print('The default user agent includes SDK version, Python version, and platform information')
    
    return client


def create_client_with_custom_user_agent():
    """
    Create a Cosmos DB client with a custom user agent suffix.
    
    The user_agent parameter appends a custom string to the default user agent,
    allowing you to identify your specific application or service.
    """
    print('\n2. Creating client with custom user agent')
    
    # Add a custom suffix to identify your application
    custom_user_agent = "MyApplication/1.0.0"
    
    client = cosmos_client.CosmosClient(
        HOST,
        {'masterKey': MASTER_KEY},
        user_agent=custom_user_agent
    )
    
    print(f'Client created with custom user agent: {custom_user_agent}')
    print('This will appear as: [default-user-agent] {}'.format(custom_user_agent))
    
    return client


def create_client_with_detailed_user_agent():
    """
    Create a client with a detailed custom user agent including multiple identifiers.
    
    You can include multiple pieces of information to make diagnostics more useful:
    - Application name and version
    - Service or microservice name
    - Environment (dev, staging, production)
    - Instance or process identifier
    """
    print('\n3. Creating client with detailed custom user agent')
    
    # Include detailed information about your application
    app_name = "OrderProcessingService"
    app_version = "2.3.1"
    environment = "production"
    instance_id = "instance-42"
    
    detailed_user_agent = f"{app_name}/{app_version} env:{environment} {instance_id}"
    
    client = cosmos_client.CosmosClient(
        HOST,
        {'masterKey': MASTER_KEY},
        user_agent=detailed_user_agent
    )
    
    print(f'Client created with detailed user agent: {detailed_user_agent}')
    print('This helps identify specific instances in diagnostics and logs')
    
    return client


def create_multiple_clients_with_different_user_agents():
    """
    Demonstrate creating multiple clients with different user agents.
    
    This is useful when you have multiple services or components in your application
    that need to be tracked separately in diagnostics.
    """
    print('\n4. Creating multiple clients with different user agents')
    
    # Client for data ingestion service
    ingestion_client = cosmos_client.CosmosClient(
        HOST,
        {'masterKey': MASTER_KEY},
        user_agent="DataIngestionService/1.0.0"
    )
    print('Created client for data ingestion: DataIngestionService/1.0.0')
    
    # Client for query service
    query_client = cosmos_client.CosmosClient(
        HOST,
        {'masterKey': MASTER_KEY},
        user_agent="QueryService/1.0.0"
    )
    print('Created client for queries: QueryService/1.0.0')
    
    # Client for analytics service
    analytics_client = cosmos_client.CosmosClient(
        HOST,
        {'masterKey': MASTER_KEY},
        user_agent="AnalyticsService/1.0.0"
    )
    print('Created client for analytics: AnalyticsService/1.0.0')
    
    print('\nNow you can distinguish requests from different services in diagnostics')
    
    return ingestion_client, query_client, analytics_client


def demonstrate_user_agent_in_operations(client, custom_suffix="DemoClient/1.0.0"):
    """
    Perform operations with a custom user agent and show how it helps with diagnostics.
    
    The user agent will appear in:
    - Azure Monitor logs
    - Cosmos DB diagnostic logs
    - Request headers sent to the service
    - Performance and usage analytics
    """
    print('\n5. Demonstrating user agent in operations')
    
    try:
        # Create database
        db = client.create_database_if_not_exists(id=DATABASE_ID)
        print(f'Database created/accessed with user agent: {custom_suffix}')
        
        # Create container
        container = db.create_container_if_not_exists(
            id=CONTAINER_ID,
            partition_key=PartitionKey(path='/id')
        )
        print(f'Container created/accessed with user agent: {custom_suffix}')
        
        # Create or upsert an item
        item = {
            'id': 'sample_item_1',
            'name': 'Sample Item',
            'description': 'Created with custom user agent'
        }
        container.upsert_item(body=item)
        print(f'Item created/updated with user agent: {custom_suffix}')
        
        # Query items
        items = list(container.query_items(
            query="SELECT * FROM c WHERE c.id = @id",
            parameters=[{"name": "@id", "value": "sample_item_1"}],
            enable_cross_partition_query=True
        ))
        print(f'Query executed with user agent: {custom_suffix}')
        print(f'Found {len(items)} item(s)')
        
        print('\nAll these operations will show your custom user agent in diagnostics')
        
        # Cleanup: Delete the item
        container.delete_item(item='sample_item_1', partition_key='sample_item_1')
        print(f'\nCleanup: Deleted sample item')
        
    except Exception as e:
        print(f'Error during operations: {e}')

def demonstrate_operations_with_multiple_user_agents():
    """
    Show how different user agents help track multiple operations.
    """
    print('\n6. Demonstrating Multiple operations with different user agents')

    def operation_with_user_agent(user_agent_suffix, operation_name):
        """Helper function to perform operation with specific user agent."""
        client = cosmos_client.CosmosClient(
                HOST,
                {'masterKey': MASTER_KEY},
                user_agent=user_agent_suffix
        )
        db = client.create_database_if_not_exists(id=DATABASE_ID)
        print(f'{operation_name} completed with user agent: {user_agent_suffix}')

    # Run multiple operations with different user agents
    operation_with_user_agent("Worker1/1.0.0", "Worker 1")
    operation_with_user_agent("Worker2/1.0.0", "Worker 2")
    operation_with_user_agent("Worker3/1.0.0", "Worker 3")

    print('\nEach operation can be tracked separately in diagnostics')

def run_sample():
    """
    Run all user agent examples.
    """
    print('=' * 70)
    print('Azure Cosmos DB - User Agent Management Sample')
    print('=' * 70)
    
    # Example 1: Default user agent
    default_client = create_client_with_default_user_agent()
    
    # Example 2: Simple custom user agent
    custom_client = create_client_with_custom_user_agent()
    
    # Example 3: Detailed user agent
    detailed_client = create_client_with_detailed_user_agent()
    
    # Example 4: Multiple clients with different user agents
    ingestion_client, query_client, analytics_client = create_multiple_clients_with_different_user_agents()
    
    # Example 5: Show user agent in operations
    demonstrate_user_agent_in_operations(detailed_client, "OrderProcessingService/2.3.1")

    # Example 6: Operations with different user agents
    demonstrate_operations_with_multiple_user_agents()

    print('\n' + '=' * 70)
    print('Sample completed!')
    print('=' * 70)

if __name__ == '__main__':
    run_sample()



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.
# cspell:ignore rerank reranker reranking
import json
import unittest
import asyncio

from azure.cosmos.aio import CosmosClient
import azure.cosmos.exceptions as exceptions
import pytest
from azure.identity.aio import DefaultAzureCredential

import test_config


@pytest.mark.semanticReranker
class TestSemanticRerankerAsync(unittest.TestCase):
    """Test to check async semantic reranker behavior."""
    client: CosmosClient = None
    config = test_config.TestConfig
    host = config.host
    TEST_DATABASE_ID = config.TEST_DATABASE_ID
    TEST_CONTAINER_ID = config.TEST_SINGLE_PARTITION_CONTAINER_ID
    TEST_CONTAINER_PARTITION_KEY = config.TEST_CONTAINER_PARTITION_KEY


    @classmethod
    def setUpClass(cls):
        if cls.host == '[YOUR_ENDPOINT_HERE]':
            raise Exception(
                "You must specify your Azure Cosmos account values for "
                "'host' at the top of this class to run the "
                "tests.")

    async def asyncSetUp(self):
        """Async setup for each test."""
        credential = DefaultAzureCredential()
        self.client = CosmosClient(self.host, credential, connection_verify=False)
        self.test_db = await self.client.create_database_if_not_exists(self.TEST_DATABASE_ID)
        self.test_container = await self.test_db.create_container_if_not_exists(
            self.TEST_CONTAINER_ID,
            self.TEST_CONTAINER_PARTITION_KEY
        )

    async def asyncTearDown(self):
        """Async teardown for each test."""
        try:
            await self.test_db.delete_container(self.TEST_CONTAINER_ID)
            await self.client.delete_database(self.TEST_DATABASE_ID)
        except exceptions.CosmosHttpResponseError:
            pass
        finally:
            await self.client.close()

    def test_semantic_reranker_async(self):
        """Test async semantic reranking functionality."""
        async def run_test():
            await self.asyncSetUp()
            try:
                documents = self._get_documents(document_type="string")
                results = await self.test_container.semantic_rerank(
                    reranking_context="What is the capital of France?",
                    documents=documents,
                    semantic_reranking_options={
                        "return_documents": True,
                        "top_k": 10,
                        "batch_size": 32,
                        "sort": True
                    }
                )
                assert len(results["Scores"]) == len(documents)
                assert results["Scores"][0]["document"] == "Paris is the capital of France."

            finally:
                await self.asyncTearDown()
        asyncio.run(run_test())

    def test_semantic_reranker_async_json_documents(self):
        async def run_test():
            await self.asyncSetUp()
            try:
                documents = self._get_documents(document_type="json")
                results = await self.test_container.semantic_rerank(
                    reranking_context="What is the capital of France?",
                    documents=[json.dumps(item) for item in documents],
                    semantic_reranking_options={
                        "return_documents": True,
                        "top_k": 10,
                        "batch_size": 32,
                        "sort": True,
                        "document_type": "json",
                        "target_paths": "text",
                    }
                )

                assert len(results["Scores"]) == len(documents)
                returned_document = json.loads(results["Scores"][0]["document"])
                assert returned_document["text"] == "Paris is the capital of France."
            finally:
                await self.asyncTearDown()
        asyncio.run(run_test())

    def test_semantic_reranker_async_nested_json_documents(self):
        async def run_test():
            await self.asyncSetUp()
            try:
                documents = self._get_documents(document_type="nested_json")
                results = await self.test_container.semantic_rerank(
                    reranking_context="What is the capital of France?",
                    documents=[json.dumps(item) for item in documents],
                    semantic_reranking_options={
                        "return_documents": True,
                        "top_k": 10,
                        "batch_size": 32,
                        "sort": True,
                        "document_type": "json",
                        "target_paths": "info.text",
                    }
                )

                assert len(results["Scores"]) == len(documents)
                returned_document = json.loads(results["Scores"][0]["document"])
                assert returned_document["info"]["text"] == "Paris is the capital of France."
            finally:
                await self.asyncTearDown()
        asyncio.run(run_test())

    def _get_documents(self, document_type: str):
        if document_type == "string":
            return [
                "Berlin is the capital of Germany.",
                "Paris is the capital of France.",
                "Madrid is the capital of Spain."
            ]
        elif document_type == "json":
            return [
                {"id": "1", "text": "Berlin is the capital of Germany."},
                {"id": "2", "text": "Paris is the capital of France."},
                {"id": "3", "text": "Madrid is the capital of Spain."}
            ]
        elif document_type == "nested_json":
            return [
                {"id": "1", "info": {"text": "Berlin is the capital of Germany."}},
                {"id": "2", "info": {"text": "Paris is the capital of France."}},
                {"id": "3", "info": {"text": "Madrid is the capital of Spain."}}
            ]
        else:
            raise ValueError("Unsupported document type")



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

import unittest
import uuid
import pytest

import azure.cosmos.cosmos_client as cosmos_client
import test_config
from azure.cosmos import DatabaseProxy
from azure.cosmos.partition_key import PartitionKey
import azure.cosmos.exceptions as exceptions

@pytest.mark.cosmosQuery
class TestComputedPropertiesQuery(unittest.TestCase):
    """Test to ensure escaping of non-ascii characters from partition key"""

    created_db: DatabaseProxy = None
    client: cosmos_client.CosmosClient = None
    config = test_config.TestConfig
    host = config.host
    masterKey = config.masterKey
    connectionPolicy = config.connectionPolicy
    TEST_DATABASE_ID = config.TEST_DATABASE_ID

    @classmethod
    def setUpClass(cls):
        if (cls.masterKey == '[YOUR_KEY_HERE]' or
                cls.host == '[YOUR_ENDPOINT_HERE]'):
            raise Exception(
                "You must specify your Azure Cosmos account values for "
                "'masterKey' and 'host' at the top of this class to run the "
                "tests.")

        cls.client = cosmos_client.CosmosClient(cls.host, cls.masterKey)
        cls.client.create_database_if_not_exists(cls.TEST_DATABASE_ID)
        cls.created_db = cls.client.get_database_client(cls.TEST_DATABASE_ID)
        cls.items = [
            {'id': str(uuid.uuid4()), 'pk': 'test', 'val': 5, 'stringProperty': 'prefixOne', 'db_group': 'GroUp1'},
            {'id': str(uuid.uuid4()), 'pk': 'test', 'val': 5, 'stringProperty': 'prefixTwo', 'db_group': 'GrOUp1'},
            {'id': str(uuid.uuid4()), 'pk': 'test', 'val': 5, 'stringProperty': 'randomWord1', 'db_group': 'GroUp2'},
            {'id': str(uuid.uuid4()), 'pk': 'test', 'val': 5, 'stringProperty': 'randomWord2', 'db_group': 'groUp1'},
            {'id': str(uuid.uuid4()), 'pk': 'test', 'val': 5, 'stringProperty': 'randomWord3', 'db_group': 'GroUp3'},
            {'id': str(uuid.uuid4()), 'pk': 'test', 'val': 5, 'stringProperty': 'randomWord4', 'db_group': 'GrOUP1'},
            {'id': str(uuid.uuid4()), 'pk': 'test', 'val': 5, 'stringProperty': 'randomWord5', 'db_group': 'GroUp2'},
            {'id': str(uuid.uuid4()), 'pk': 'test', 'val': 0, 'stringProperty': 'randomWord6', 'db_group': 'group1'},
            {'id': str(uuid.uuid4()), 'pk': 'test', 'val': 3, 'stringProperty': 'randomWord7', 'db_group': 'group2'},
            {'id': str(uuid.uuid4()), 'pk': 'test', 'val': 2, 'stringProperty': 'randomWord8', 'db_group': 'GroUp3'}
        ]
        cls.computed_properties = [{'name': "cp_lower", 'query': "SELECT VALUE LOWER(c.db_group) FROM c"},
                                   {'name': "cp_power", 'query': "SELECT VALUE POWER(c.val, 2) FROM c"},
                                   {'name': "cp_str_len", 'query': "SELECT VALUE LENGTH(c.stringProperty) FROM c"}]

    def computedPropertiesTestCases(self, created_collection):
        # Check that computed properties were properly sent
        self.assertListEqual(self.computed_properties, created_collection.read()["computedProperties"])

        # Test 0: Negative test, test if using non-existent computed property
        queried_items = list(
            created_collection.query_items(query='Select * from c Where c.cp_upper = "GROUP2"',
                                           partition_key="test"))
        self.assertEqual(len(queried_items), 0)

        # Test 1: Test first computed property
        queried_items = list(
            created_collection.query_items(query='Select * from c Where c.cp_lower = "group1"', partition_key="test"))
        self.assertEqual(len(queried_items), 5)

        # Test 1 Negative: Test if using non-existent string in group property returns nothing
        queried_items = list(
            created_collection.query_items(query='Select * from c Where c.cp_lower = "group4"', partition_key="test"))
        self.assertEqual(len(queried_items), 0)

        # Test 2: Test second computed property
        queried_items = list(
            created_collection.query_items(query='Select * from c Where c.cp_power = 25', partition_key="test"))
        self.assertEqual(len(queried_items), 7)

        # Test 2 Negative: Test Non-Existent POWER
        queried_items = list(
            created_collection.query_items(query='Select * from c Where c.cp_power = 16', partition_key="test"))
        self.assertEqual(len(queried_items), 0)

        # Test 3: Test Third Computed Property
        queried_items = list(
            created_collection.query_items(query='Select * from c Where c.cp_str_len = 9', partition_key="test"))
        self.assertEqual(len(queried_items), 2)

        # Test 3 Negative: Test Str length that isn't there
        queried_items = list(
            created_collection.query_items(query='Select * from c Where c.cp_str_len = 3', partition_key="test"))
        self.assertEqual(len(queried_items), 0)

    def test_computed_properties_query(self):

        created_collection = self.created_db.create_container(
            "computed_properties_query_test_" + str(uuid.uuid4()),
            PartitionKey(path="/pk"),
            computed_properties=self.computed_properties)

        # Create Items
        for item in self.items:
            created_collection.create_item(body=item)

        self.computedPropertiesTestCases(created_collection)
        self.created_db.delete_container(created_collection.id)

    def test_replace_with_same_computed_properties(self):
        created_collection = self.created_db.create_container(
            id="computed_properties_query_test_" + str(uuid.uuid4()),
            partition_key=PartitionKey(path="/pk"),
            computed_properties=self.computed_properties)

        # Create Items
        for item in self.items:
            created_collection.create_item(body=item)
        # Check that computed properties were properly sent
        self.assertListEqual(self.computed_properties, created_collection.read()["computedProperties"])

        replaced_collection= self.created_db.replace_container(
            container=created_collection.id,
            partition_key=PartitionKey(path="/pk"),
            computed_properties= self.computed_properties)

        self.computedPropertiesTestCases(replaced_collection)
        self.created_db.delete_container(replaced_collection.id)

    def test_replace_without_computed_properties(self):
        created_collection = self.created_db.create_container(
            id="computed_properties_query_test_" + str(uuid.uuid4()),
            partition_key=PartitionKey(path="/pk"))

        # Create Items
        for item in self.items:
            created_collection.create_item(body=item)

        # Replace Container
        replaced_collection= self.created_db.replace_container(
            container=created_collection.id,
            partition_key=PartitionKey(path="/pk"),
            computed_properties= self.computed_properties
        )
        self.computedPropertiesTestCases(replaced_collection)
        self.created_db.delete_container(replaced_collection.id)

    def test_replace_with_new_computed_properties(self):
        created_collection = self.created_db.create_container(
            id="computed_properties_query_test_" + str(uuid.uuid4()),
            partition_key=PartitionKey(path="/pk"),
            computed_properties=self.computed_properties)

        # Create Items
        for item in self.items:
            created_collection.create_item(body=item)

        # Check that computed properties were properly sent
        self.assertListEqual(self.computed_properties, created_collection.read()["computedProperties"])

        new_computed_properties = [{'name': "cp_upper", 'query': "SELECT VALUE UPPER(c.db_group) FROM c"},
                                   {'name': "cp_len", 'query': "SELECT VALUE LENGTH(c.stringProperty) FROM c"}]
        # Replace Container
        replaced_collection = self.created_db.replace_container(
            container=created_collection.id,
            partition_key=PartitionKey(path="/pk"),
            computed_properties=new_computed_properties
        )
        # Check that computed properties were properly sent to replaced container
        self.assertListEqual(new_computed_properties, replaced_collection.read()["computedProperties"])

        # Test 1: Test first computed property
        queried_items = list(
            replaced_collection.query_items(query='Select * from c Where c.cp_upper = "GROUP2"',
                                            partition_key="test"))
        self.assertEqual(len(queried_items), 3)

        # Test 1 Negative: Test if using non-existent computed property name returns nothing
        queried_items = list(
            replaced_collection.query_items(query='Select * from c Where c.cp_lower = "group1"', partition_key="test"))
        self.assertEqual(len(queried_items), 0)

        # Test 2: Test Second Computed Property
        queried_items = list(
            replaced_collection.query_items(query='Select * from c Where c.cp_len = 9', partition_key="test"))
        self.assertEqual(len(queried_items), 2)

        # Test 2 Negative: Test Str length using old computed properties name
        queried_items = list(
            replaced_collection.query_items(query='Select * from c Where c.cp_str_len = 9', partition_key="test"))
        self.assertEqual(len(queried_items), 0)
        self.created_db.delete_container(created_collection.id)

    def test_replace_with_incorrect_computed_properties(self):
        created_collection = self.created_db.create_container(
            id="computed_properties_query_test_" + str(uuid.uuid4()),
            partition_key=PartitionKey(path="/pk"))

        # Create Items
        for item in self.items:
            created_collection.create_item(body=item)

        computed_properties = {'name': "cp_lower", 'query': "SELECT VALUE LOWER(c.db_group) FROM c"}

        try:
            # Replace Container with wrong type for computed_properties
            self.created_db.replace_container(
                container=created_collection.id,
                partition_key=PartitionKey(path="/pk"),
                computed_properties= computed_properties
            )
            pytest.fail("Container creation should have failed for invalid input.")
        except exceptions.CosmosHttpResponseError as e:
            assert e.status_code == 400
            assert "One of the specified inputs is invalid" in e.http_error_message

    def test_replace_with_remove_computed_properties_(self):
        created_collection = self.created_db.create_container(
            "computed_properties_query_test_" + str(uuid.uuid4()),
            PartitionKey(path="/pk"),
            computed_properties=self.computed_properties)

        # Create Items
        for item in self.items:
            created_collection.create_item(body=item)

        # Check if computed properties were set
        container = created_collection.read()
        assert self.computed_properties == container["computedProperties"]

        # Replace Container
        replaced_collection = self.created_db.replace_container(
            container=created_collection.id,
            partition_key=PartitionKey(path="/pk"))

        # Check if computed properties were not set
        container = replaced_collection.read()

        # If keyError is not raised the test will fail
        with pytest.raises(KeyError):
            computed_properties = container["computedProperties"]

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

import unittest
import uuid

import pytest

import azure.cosmos.cosmos_client as cosmos_client
import test_config
from azure.cosmos import DatabaseProxy, ContainerProxy


@pytest.mark.cosmosEmulator
class TestEncoding(unittest.TestCase):
    """Test to ensure escaping of non-ascii characters from partition key"""

    host = test_config.TestConfig.host
    masterKey = test_config.TestConfig.masterKey
    connectionPolicy = test_config.TestConfig.connectionPolicy
    client: cosmos_client.CosmosClient = None
    created_db: DatabaseProxy = None
    created_container: ContainerProxy = None

    @classmethod
    def setUpClass(cls):
        if (cls.masterKey == '[YOUR_KEY_HERE]' or
                cls.host == '[YOUR_ENDPOINT_HERE]'):
            raise Exception(
                "You must specify your Azure Cosmos account values for "
                "'masterKey' and 'host' at the top of this class to run the "
                "tests.")

        cls.client = cosmos_client.CosmosClient(cls.host, cls.masterKey)
        cls.created_db = cls.client.get_database_client(test_config.TestConfig.TEST_DATABASE_ID)
        cls.created_container = cls.created_db.get_container_client(
            test_config.TestConfig.TEST_SINGLE_PARTITION_CONTAINER_ID)

    def test_unicode_characters_in_partition_key(self):
        test_string = u'€€ کلید پارتیشن विभाजन कुंजी 	123'  # cspell:disable-line
        document_definition = {'pk': test_string, 'id': 'myid' + str(uuid.uuid4())}
        created_doc = self.created_container.create_item(body=document_definition)

        read_doc = self.created_container.read_item(item=created_doc['id'], partition_key=test_string)
        self.assertEqual(read_doc['pk'], test_string)

    def test_create_document_with_line_separator_para_seperator_next_line_unicodes(self):
        test_string = u'Line Separator ( ) & Paragraph Separator ( ) & Next Line () & نیم‌فاصله'  # cspell:disable-line
        document_definition = {'pk': 'pk', 'id': 'myid' + str(uuid.uuid4()), 'unicode_content': test_string}
        created_doc = self.created_container.create_item(body=document_definition)

        read_doc = self.created_container.read_item(item=created_doc['id'], partition_key='pk')
        self.assertEqual(read_doc['unicode_content'], test_string)

    def test_create_stored_procedure_with_line_separator_para_seperator_next_line_unicodes(self):
        test_string = 'Line Separator ( ) & Paragraph Separator ( ) & Next Line () & نیم‌فاصله'  # cspell:disable-line

        test_string_unicode = u'Line Separator ( ) & Paragraph Separator ( ) & Next Line () & نیم‌فاصله'  # cspell:disable-line

        stored_proc_definition = {'id': 'myid' + str(uuid.uuid4()), 'body': test_string}
        created_sp = self.created_container.scripts.create_stored_procedure(body=stored_proc_definition)

        read_sp = self.created_container.scripts.get_stored_procedure(created_sp['id'])
        self.assertEqual(read_sp['body'], test_string_unicode)


if __name__ == "__main__":
    unittest.main()
